# This module implements the base classes for proteins and
# nucleic acid chains.
#
# Written by Konrad Hinsen
# last revision: 1999-3-7
#

import Bonds, ChemicalObjects, Collection, Database, PDB, Units
from Scientific.DictWithDefault import DictWithDefault
from Scientific.Geometry import Objects3D
import umath, string

class Residue(ChemicalObjects.Group):

    def __setstate__(self, state):
	for attr, value in state.items():
	    self.__dict__[attr] = value
	try:
	    self.model = self.hydrogens
	except AttributeError:
	    pass
	self._init()

    def _init(self):
	# construct PDB map and alternative names
	type = self.type
	if not hasattr(type, 'pdbmap'):
	    pdb_dict = {}
	    type.pdbmap = [(type.symbol, pdb_dict)]
	    offset = 0
	    for g in type.groups:
		for name, atom in g.pdbmap[0][1].items():
		    pdb_dict[name] = Database.AtomReference(atom.number + \
							    offset)
		offset = offset + len(g.atoms)
	if not hasattr(type, 'pdb_alternative'):
	    alt_dict = {}
	    for g in type.groups:
		if hasattr(g, 'pdb_alternative'):
		    for key, value in g.pdb_alternative.items():
			alt_dict[key] = value
	    setattr(type, 'pdb_alternative', alt_dict)


class ResidueChain(ChemicalObjects.Molecule):

    is_chain = 1

    def _setupChain(self, circular, properties, conf):
        self.atoms = []
        self.bonds = []
        for g in self.groups:
            self.atoms = self.atoms + g.atoms
            self.bonds = self.bonds + g.bonds
        for i in range(len(self.groups)-1):
            link1 = self.groups[i].chain_links[1]
            link2 = self.groups[i+1].chain_links[0]
            self.bonds.append(Bonds.Bond((link1, link2)))
        if circular:
            link1 = self.groups[-1].chain_links[1]
            link2 = self.groups[0].chain_links[0]
            self.bonds.append(Bonds.Bond((link1, link2)))
        self.bonds = Bonds.BondList(self.bonds)
        self.parent = None
        self.type = None
        self.configurations = {}
        try:
            self.name = properties['name']
            del properties['name']
        except KeyError:
            self.name = ''
        if conf:
            conf.applyConfiguration(self)
        if properties.has_key('position'):
            self.translateTo(properties['position'])
            del properties['position']
        self.addProperties(properties)

    def __len__(self):
	return len(self.groups)

    def __getitem__(self, item):
	return self.groups[item]

    def __setitem__(self, item, value):
	self.replaceResidue(self.groups[item], value)

    def residuesOfType(self, *types):
	types = map(string.lower, types)
	rlist = []
	for r in self.groups:
	    if string.lower(r.type.symbol) in types:
		rlist.append(r)
	return Collection.Collection(rlist)

    def residues(self):
	return Collection.Collection(self.groups)

    def sequence(self):
	return map(lambda r: string.lower(r.type.symbol), self.groups)

    # construct positions of missing hydrogens
    def findHydrogenPositions(self):
	self.setBondAttributes()
	try:
	    unknown = DictWithDefault([])
	    for a in self.atoms:
		if a.position() is None:
		    if a.symbol != 'H':
			raise ValueError, 'position of ' + a.fullName() + \
			      ' is undefined'
		    bonded = a.bondedTo()[0]
		    unknown[bonded].append(a)
	    for a, list in unknown.items():
		bonded = a.bondedTo()
		n = len(bonded)
		known = []
		for b in bonded:
		    if b.position() is not None:
			known.append(b)
		nb = len(list)
		if a.symbol == 'C':
		    if n == 4:
			if nb == 1:
			    self._C4oneH(a, known, list)
			elif nb == 2:
			    self._C4twoH(a, known, list)
			elif nb == 3:
			    self._C4threeH(a, known, list)
		    elif n == 3:
			self._C3oneH(a, known, list)
		    else:
			print a
			raise ValueError, "Can't handle C with "+`n`+" bonds"
		elif a.symbol == 'N':
		    if n == 4:
			if nb == 3:
			    self._N4threeH(a, known, list)
		    elif n == 3:
			if nb == 1:
			    self._N3oneH(a, known, list)
			elif nb == 2:
			    self._N3twoH(a, known, list)
		    else:
			print a
			raise ValueError, "Can't handle N with "+`n`+" bonds"
		elif a.symbol == 'O' and n == 2:
		    self._O2(a, known, list)
		elif a.symbol == 'S' and n == 2:
		    self._S2(a, known, list)
		else:
		    print a
		    raise ValueError, "Can't handle this yet: " + \
			  a.symbol + ' with ' + `n` + ' bonds (' + \
			  a.fullName() + ').'
	finally:
	    self.clearBondAttributes()

    # default C-H bond length and X-C-H angle
    _ch_bond = 1.09*Units.Ang
    _hch_angle = umath.arccos(-1./3.)*Units.rad
    _nh_bond = 1.03*Units.Ang
    _hnh_angle = 120.*Units.deg
    _oh_bond = 0.95*Units.Ang
    _coh_angle = 114.9*Units.deg
    _sh_bond = 1.007*Units.Ang
    _csh_angle = 96.5*Units.deg

    def _C4oneH(self, atom, known, unknown):
	r = atom.position()
	n0 = (known[0].position()-r).normal()
	n1 = (known[1].position()-r).normal()
	n2 = (known[2].position()-r).normal()
	n3 = (n0 + n1 + n2).normal()
	unknown[0].setPosition(r-self._ch_bond*n3)

    def _C4twoH(self, atom, known, unknown):
	r = atom.position()
	r1 = known[0].position()
	r2 = known[1].position()
	plane = Objects3D.Plane(r, r1, r2)
	axis = -((r1-r)+(r2-r)).normal()
	plane = plane.rotate(Objects3D.Line(r, axis), 90.*Units.deg)
	cone = Objects3D.Cone(r, axis, 0.5*self._hch_angle)
	sphere = Objects3D.Sphere(r, self._ch_bond)
	circle = sphere.intersectWith(cone)
	points = circle.intersectWith(plane)
	unknown[0].setPosition(points[0])
	unknown[1].setPosition(points[1])

    def _C4threeH(self, atom, known, unknown):
	self._tetrahedralH(atom, known, unknown, self._ch_bond)

    def _C3oneH(self, atom, known, unknown):
	r = atom.position()
	n1 = (known[0].position()-r).normal()
	n2 = (known[1].position()-r).normal()
	n3 = -(n1 + n2).normal()
	unknown[0].setPosition(r+self._ch_bond*n3)

    def _N3oneH(self, atom, known, unknown):
	r = atom.position()
	n1 = (known[0].position()-r).normal()
	n2 = (known[1].position()-r).normal()
	n3 = -(n1 + n2).normal()
	unknown[0].setPosition(r+self._nh_bond*n3)

    def _N3twoH(self, atom, known, unknown):
	r = atom.position()
	r1 = known[0].position()
	others = filter(lambda a: a.symbol != 'H', known[0].bondedTo())
	r2 = others[0].position()
	plane = Objects3D.Plane(r, r1, r2)
	axis = (r-r1).normal()
	cone = Objects3D.Cone(r, axis, 0.5*self._hnh_angle)
	sphere = Objects3D.Sphere(r, self._nh_bond)
	circle = sphere.intersectWith(cone)
	points = circle.intersectWith(plane)
	unknown[0].setPosition(points[0])
	unknown[1].setPosition(points[1])

    def _N4threeH(self, atom, known, unknown):
	self._tetrahedralH(atom, known, unknown, self._nh_bond)

    def _O2(self, atom, known, unknown):
	others = known[0].bondedTo()
        for a in others:
            r = a.position()
            if a != atom and r is not None: break
	dihedral = 180.*Units.deg
	self._findPosition(unknown[0], atom.position(), known[0].position(), r,
			   self._oh_bond, self._coh_angle, dihedral)

    def _S2(self, atom, known, unknown):
	c2 = filter(lambda a: a.symbol == 'C', known[0].bondedTo())[0]
	self._findPosition(unknown[0], atom.position(), known[0].position(),
			   c2.position(),
			   self._sh_bond, self._csh_angle,
			   180.*Units.deg)

    def _tetrahedralH(self, atom, known, unknown, bond):
	r = atom.position()
	n = (known[0].position()-r).normal()
	cone = Objects3D.Cone(r, n, umath.arccos(-1./3.))
	sphere = Objects3D.Sphere(r, bond)
	circle = sphere.intersectWith(cone)
	others = filter(lambda a: a.symbol != 'H', known[0].bondedTo())
	others.remove(atom)
	other = others[0]
	ref = (Objects3D.Plane(circle.center, circle.normal) \
	       .projectionOf(other.position())-circle.center).normal()
	p0 = circle.center + circle.radius*ref
	p0 = Objects3D.rotatePoint(p0,
				  Objects3D.Line(circle.center, circle.normal),
				  60.*Units.deg)
	p1 = Objects3D.rotatePoint(p0,
				  Objects3D.Line(circle.center, circle.normal),
				  120.*Units.deg)
	p2 = Objects3D.rotatePoint(p1,
				  Objects3D.Line(circle.center, circle.normal),
				  120.*Units.deg)
	unknown[0].setPosition(p0)
	unknown[1].setPosition(p1)
	unknown[2].setPosition(p2)

    def _findPosition(self, unknown, a1, a2, a3, bond, angle, dihedral):
	sphere = Objects3D.Sphere(a1, bond)
	cone = Objects3D.Cone(a1, a2-a1, angle)
	plane = Objects3D.Plane(a3, a2, a1)
	plane = plane.rotate(Objects3D.Line(a1, a2-a1), dihedral)
	points = sphere.intersectWith(cone).intersectWith(plane)
	for p in points:
	    if (a1-a2).cross(p-a1)*(plane.normal) > 0:
		unknown.setPosition(p)
		break

#
# Find the full name of a residue
#
def _fullName(residue):
    return _residue_names[string.lower(residue)]

_residue_names = {'ala': 'alanine',          'a': 'alanine',
		  'arg': 'arginine',         'r': 'arginine',
		  'asn': 'asparagine',       'n': 'asparagine',
		  'asp': 'aspartic_acid',    'd': 'aspartic_acid',
		  'cys': 'cysteine',         'c': 'cysteine',
		  'gln': 'glutamine',        'q': 'glutamine',
		  'glu': 'glutamic_acid',    'e': 'glutamic_acid',
		  'gly': 'glycine',          'g': 'glycine',
		  'his': 'histidine',        'h': 'histidine',
		  'ile': 'isoleucine',       'i': 'isoleucine',
		  'leu': 'leucine',          'l': 'leucine',
		  'lys': 'lysine',           'k': 'lysine',
		  'met': 'methionine',       'm': 'methionine',
		  'phe': 'phenylalanine',    'f': 'phenylalanine',
		  'pro': 'proline',          'p': 'proline',
		  'ser': 'serine',           's': 'serine',
		  'thr': 'threonine',        't': 'threonine',
		  'trp': 'tryptophan',       'w': 'tryptophan',
		  'tyr': 'tyrosine',         'y': 'tyrosine',
		  'val': 'valine',           'v': 'valine',
		  'cyx': 'cystine_ss',
		  'hsd': 'histidine_deltah', 'hse': 'histidine_epsilonh',
		  'hsp': 'histidine_plus',
		  'ace': 'ace_beginning',    'nme': 'nmethyl',

                  'da': 'd-adenosine',
                  'dc': 'd-cytosine',
                  'dg': 'd-guanosine',
                  'dt': 'd-thymine',
                  'ra': 'r-adenosine',
                  'rc': 'r-cytosine',
                  'rg': 'r-guanosine',
                  'ru': 'r-uracil'

                  }

#
# Add a residue to the residue list
#
def defineAminoAcidResidue(full_name, code3, code1 = None):
    if _residue_names.has_key(code3):
	raise ValueError, "residue name " + code3 + " already used"
    if _residue_names.has_key(code1):
	raise ValueError, "residue name " + code1 + " already used"
    _residue_names[code3] = full_name
    if code1 is not None:
	_residue_names[code1] = full_name
    import Scientific.IO.PDB
    Scientific.IO.PDB.defineAminoAcidResidue(residue.symbol)

def defineNucleicAcidResidue(full_name, code):
    if _residue_names.has_key(code):
	raise ValueError, "residue name " + code + " already used"
    _residue_names[code] = full_name
    import Scientific.IO.PDB
    Scientific.IO.PDB.defineNucleicAcidResidue(residue.symbol)
