# This module handles input and application of Amber parameter files.
#
# Written by Konrad Hinsen
# last revision: 1999-2-25
#

from MMTK import Units
from Scientific.IO.FortranFormat import FortranFormat, FortranLine
from Scientific.IO.TextFile import TextFile
from Scientific.DictWithDefault import DictWithDefault
import os, string

amber_energy_unit = Units.kcal/Units.mol
amber_length_unit = Units.Ang
amber_angle_unit = Units.deg
amber_fc_angle_unit = Units.rad

bond_order = { 'HC' : 1, 'HA' : 1, 'HA' : 1, 'H1' : 1, 'H2' : 1, 'H3' : 1,\
	       'H4' : 1, 'SI' : 4, \
	       'CT' : 4, 'CA' : 3, 'CM' : 3, 'C' : 3,\
	       'O' : 2, 'OS' : 2, 'O2' : 1, 'OH' : 2,\
	       'NC' : 3, 'N2' : 3, 'N*' : 3, 'P' : 4 }

#
# The force field parameter file
#
class AmberParameters:

    def __init__(self, filename):
	file = TextFile(filename)
	title = file.readline()[:-1]

	format = FortranFormat('A2,1X,F10.2')
	self.atom_types = DictWithDefault(None)
	while 1:
	    l = FortranLine(file.readline()[:-1], format)
	    if l.isBlank(): break
	    name = _normalizeName(l[0])
	    self.atom_types[name] = AmberAtomType(name, l[1])

	format = FortranFormat('20(A2,2X)')
	done = 0
	while not done:
	    l = FortranLine(file.readline()[:-1], format)
	    for entry in l:
		name = _normalizeName(entry)
		if len(name) == 0:
		    done = 1
		    break
		try: # ignore errors for now
		    self.atom_types[name].hydrophylic = 1
		except: pass

	format = FortranFormat('A2,1X,A2,2F10.2')
	self.bonds = {}
	while 1:
	    l = FortranLine(file.readline()[:-1], format)
	    if l.isBlank(): break
	    name1 = _normalizeName(l[0])
	    name2 = _normalizeName(l[1])
	    name1, name2 = _sort(name1, name2)
	    self.bonds[(name1, name2)] = \
			       AmberBondParameters(self.atom_types[name1],
						   self.atom_types[name2],
						   l[2], l[3])

	format = FortranFormat('A2,1X,A2,1X,A2,2F10.2')
	self.bond_angles = {}
	self.bond_angles_average = {}
	bond_angle_sum = {}
	while 1:
	    l = FortranLine(file.readline()[:-1], format)
	    if l.isBlank(): break
	    name1 = _normalizeName(l[0])
	    name2 = _normalizeName(l[1])
	    name3 = _normalizeName(l[2])
	    name1, name3 = _sort(name1, name3)
	    p = AmberBondAngleParameters(self.atom_types[name1],
					 self.atom_types[name2],
					 self.atom_types[name3],
					 l[3], l[4])
	    self.bond_angles[(name1, name2, name3)] = p
	    try:
		(asum,ksum,count) = bond_angle_sum[(name1, name2)]
	    except:
		asum = 0
		ksum = 0
		count = 0
	    bond_angle_sum[(name1, name2)] = (asum + l[4],
					      ksum + l[3], count + 1)
	    try:
		(asum,ksum,count) = bond_angle_sum[name2]
	    except:
		asum = 0
		ksum = 0
		count = 0
	    bond_angle_sum[name2] = (asum + l[4], ksum + l[3], count + 1)
	for key in bond_angle_sum.keys():
	    (angle_sum, k_sum, n) = bond_angle_sum[key]
	    try:
		(k1, k2) = key
	    except:
		k1 = None
		k2 = key
	    self.bond_angles_average[key] =\
		AmberBondAngleParameters(k1, k2, None, k_sum/n, angle_sum/n)

	format = FortranFormat('A2,1X,A2,1X,A2,1X,A2,I4,3F15.2')
	self.dihedrals = {}
	self.dihedrals_2 = {}
	append = None
	while 1:
	    l = FortranLine(file.readline()[:-1], format)
	    if l.isBlank(): break
	    name1 = _normalizeName(l[0])
	    name2 = _normalizeName(l[1])
	    name3 = _normalizeName(l[2])
	    name4 = _normalizeName(l[3])
	    name1, name2, name3, name4 = _sort4(name1, name2, name3, name4)
	    if append is not None:
		append.addTerm(l[4], l[5], l[6], l[7])
		if l[7] >= 0: append = None
	    else:
		p = AmberDihedralParameters(self.atom_types[name1],
					    self.atom_types[name2],
					    self.atom_types[name3],
					    self.atom_types[name4],
					    l[4], l[5], l[6], l[7])
		if name1 == 'X' and name4 == 'X':
		    self.dihedrals_2[(name2, name3)] = p
		else:
		    self.dihedrals[(name1, name2, name3, name4)] = p
		if l[7] < 0: append = p

	format = FortranFormat('A2,1X,A2,1X,A2,1X,A2,I4,3F15.2')
	self.impropers = {}
	self.impropers_1 = {}
	self.impropers_2 = {}
	while 1:
	    l = FortranLine(file.readline()[:-1], format)
	    if l.isBlank(): break
	    name1 = _normalizeName(l[0])
	    name2 = _normalizeName(l[1])
	    name3 = _normalizeName(l[2])
	    name4 = _normalizeName(l[3])
	    if name1 == 'X':
		if name2 == 'X':
		    name1 = name3
		    name2 = name4
		    name3 = 'X'
		    name4 = 'X'
		else:
		    name1 = name3
		    name2, name3 = _sort(name2, name4)
		    name4 = 'X'
	    else:
		name1, name2, name3, name4 = \
		       (name3, ) + _sort3(name1, name2, name4)
	    p = AmberImproperParameters(self.atom_types[name1],
					self.atom_types[name2],
					self.atom_types[name3],
					self.atom_types[name4],
					l[4], l[5], l[6], l[7])
	    if name4 == 'X':
		if name3 == 'X':
		    self.impropers_2[(name1, name2)] = p
		else:
		    self.impropers_1[(name1, name2, name3)] = p
	    else:
		self.impropers[(name1, name2, name3, name4)] = p

	format = FortranFormat('2X,A2,2X,A2,2X,2F10.2')
	self.hbonds = {}
	while 1:
	    l = FortranLine(file.readline()[:-1], format)
	    if l.isBlank(): break
	    name1 = _normalizeName(l[0])
	    name2 = _normalizeName(l[1])
	    name1, name2 = _sort(name1, name2)
	    self.hbonds[(name1, name2)] = \
				AmberHbondParameters(self.atom_types[name1],
						     self.atom_types[name2],
						     l[2], l[3])

	format = FortranFormat('20(A2,2X)')
	self.lj_equivalent = {}
	while 1:
	    l = FortranLine(file.readline()[:-1], format)
	    if l.isBlank(): break
	    name1 = _normalizeName(l[0])
	    for s in l[1:]:
		name2 = _normalizeName(s)
		self.lj_equivalent[name2] = name1

	format = FortranFormat('2X,A2,6X,3F10.6')
	self.ljpar_sets = {}
	while 1:
	    l = FortranLine(file.readline()[:-1], 'A4,6X,A2')
	    if l[0] == 'END ': break
	    set_name = _normalizeName(l[0])
	    ljpar_set = AmberLJParameterSet(set_name, l[1])
	    self.ljpar_sets[set_name] = ljpar_set
	    while 1:
		l = FortranLine(file.readline()[:-1], format)
		if l.isBlank(): break
		name = _normalizeName(l[0])
		ljpar_set.addEntry(name, l[1], l[2], l[3])

	file.close()

    def bondParameters(self, name1, name2):
	name1 = _normalizeName(name1)
	name2 = _normalizeName(name2)
	name1, name2 = _sort(name1, name2)
	p = self.bonds[(name1, name2)]
	return (p.l*amber_length_unit,
		p.k*amber_energy_unit/amber_length_unit**2)

    def bondAngleParameters(self, name1, name2, name3):
	name1 = _normalizeName(name1)
	name2 = _normalizeName(name2)
	name3 = _normalizeName(name3)
	name1, name3 = _sort(name1, name3)
	p = self.bond_angles[(name1, name2, name3)]
	return (p.angle*amber_angle_unit,
		p.k*amber_energy_unit/amber_fc_angle_unit)

    def bondAngleAvgParameters(self, name1, name2):
	name1 = _normalizeName(name1)
	if name2 is None:
	    p = self.bond_angles_average[name1]
	else:
	    name2 = _normalizeName(name2)
	    p = self.bond_angles_average[(name1, name2)]
	return (p.angle*amber_angle_unit,
		p.k*amber_energy_unit/amber_fc_angle_unit)

    def dihedralParameters(self, name1, name2, name3, name4):
	name1 = _normalizeName(name1)
	name2 = _normalizeName(name2)
	name3 = _normalizeName(name3)
	name4 = _normalizeName(name4)
	name1, name2, name3, name4 = _sort4(name1, name2, name3, name4)
	try:
	    p = self.dihedrals[(name1, name2, name3, name4)]
	except KeyError:
	    try:
		p = self.dihedrals_2[(name2, name3)]
	    except KeyError: p = None
	if p is None:
	    return p
	else:
	    return map(lambda p: (p[2], p[1]*amber_angle_unit,
				  p[0]*amber_energy_unit),
		       p.terms)

    def improperParameters(self, name1, name2, name3, name4):
	name1 = _normalizeName(name1)
	name2 = _normalizeName(name2)
	name3 = _normalizeName(name3)
	name4 = _normalizeName(name4)
	name2, name3, name4 = _sort3(name2, name3, name4)
	p = None
	try:
	    p = self.impropers[(name1, name2, name3, name4)]
	except KeyError:
	    for names in [(name2, name3), (name2, name4), (name3, name4)]:
		try:
		    p = self.impropers_1[(name1,) + names]
		except KeyError: pass
	    if p is None:
		for name in [name2, name3, name4]:
		    try:
			p = self.impropers_2[(name1, name)]
		    except KeyError: pass
	if p is None:
	    return p
	else:
	    return map(lambda p: (p[2], p[1]*amber_angle_unit,
				  p[0]*amber_energy_unit),
		       p.terms)

    def ljParameters(self, name, parameter_set = None):
	if parameter_set is None:
	    ljp = self.default_ljpar_set
	else:
	    ljp = self.ljpar_sets[parameter_set]
	try:
	    name = self.lj_equivalent[name]
	except KeyError: pass
	if ljp.type == 'RE':
	    p = ljp.getEntry(name)
	    return (p[1]*amber_energy_unit,
		    1.78179743628*p[0]*amber_length_unit)
	else:
	    raise ValueError, 'Type ' + ljp.type + ' not supported.'

    def bondOrder(self, name):
	name = _normalizeName(name)
	return bond_order[name]

#
# Atom type class
#
class AmberAtomType:

    def __init__(self, name, mass):
	self.name = name
	self.mass = mass
	self.hydrophylic = 0

#
# Bond parameter class
#
class AmberBondParameters:

    def __init__(self, a1, a2, k, l):
	self.a1 = a1
	self.a2 = a2
	self.k = k
	self.l = l

#
# Bond angle parameter class
#
class AmberBondAngleParameters:

    def __init__(self, a1, a2, a3, k, angle):
	self.a1 = a1
	self.a2 = a2
	self.a3 = a3
	self.k = k
	self.angle = angle

#
# Dihedral angle parameter class
#
class AmberDihedralParameters:

    def __init__(self, a1, a2, a3, a4, divf, k, delta, n):
	self.a1 = a1
	self.a2 = a2
	self.a3 = a3
	self.a4 = a4
	self.terms = [(k/divf, delta, int(abs(n)))]

    def addTerm(self, divf, k, delta, n):
	self.terms.append((k/divf, delta, int(abs(n))))

    def __repr__(self):
	if self.a1 is None and self.a4 is None:
	    return 'X-' + self.a2.name + '-' + self.a3.name + '-X'
	else:
	    return self.a1.name + '-' + self.a2.name + '-' + \
		   self.a3.name + '-' + self.a4.name
    __str__ = __repr__

#
# Improper dihedral angle parameter class
#
class AmberImproperParameters:

    def __init__(self, a1, a2, a3, a4, divf, k, delta, n):
	self.a1 = a1
	self.a2 = a2
	self.a3 = a3
	self.a4 = a4
	self.terms = [(0.5*k, delta, int(abs(n)))]

    def addTerm(self, divf, k, delta, n):
	self.terms.append((0.5*k, delta, int(abs(n))))

    def __repr__(self):
	if self.a4 is None:
	    if self.a3 is None:
		return self.a1.name + '-' + self.a2.name + '-X-X'
	    else:
		return self.a1.name + '-' + self.a2.name + '-' + \
		       self.a3.name + '-X'
	else:
	    return self.a1.name + '-' + self.a2.name + '-' + \
		   self.a3.name + '-' + self.a4.name
    __str__ = __repr__

#
# H-bond parameter class
#
class AmberHbondParameters:

    def __init__(self, a1, a2, a, b):
	self.a1 = a1
	self.a2 = a2
	self.a = a
	self.b = b

#
# Lennard-Jones parameter sets
#
class AmberLJParameterSet:

    def __init__(self, name, type):
	self.name = name
	self.type = type
	self.entries = {}

    def addEntry(self, name, p1, p2, p3):
	if self.type == 'SK':
	    self.entries[name] = (p1, p2, p3)
	else:
	    self.entries[name] = (p1, p2)

    def getEntry(self, name):
	return self.entries[name]

#
# Utility functions
#
def _normalizeName(name):
    return string.upper(string.strip(name))

def _sort(s1, s2):
    if s2 < s1:
	return s2, s1
    else:
	return s1, s2

def _sort3(s1, s2, s3):
    if s2 < s1:
	s1, s2 = s2, s1
    if s3 < s2:
	s2, s3 = s3, s2
    if s2 < s1:
	s1, s2 = s2, s1
    return s1, s2, s3

def _sort4(s1, s2, s3, s4):
    if s3 < s2:
	return s4, s3, s2, s1
    else:
	return s1, s2, s3, s4

this_directory = os.path.split(__file__)[0]

Amber94 = AmberParameters(os.path.join(this_directory, "amber_parm94.dat"))
Amber94.lennard_jones_1_4 = 0.5
Amber94.electrostatic_1_4 = 1./1.2
Amber94.default_ljpar_set = Amber94.ljpar_sets['MOD4']
Amber94.atom_type_property = 'amber_atom_type'
Amber94.charge_property = 'amber_charge'

Amber91 = AmberParameters(os.path.join(this_directory, "amber_parm91X.dat"))
Amber91.lennard_jones_1_4 = 0.5
Amber91.electrostatic_1_4 = 0.5
Amber91.default_ljpar_set = Amber91.ljpar_sets['STDA']
Amber91.atom_type_property = 'amber91_atom_type'
Amber91.charge_property = 'amber91_charge'
