# This module implements classes that represent force fields
# for non-bonded interactions
#
# Written by Konrad Hinsen
# last revision: 1999-3-23
#

from MMTK import Units, Utility
from ForceField import ForceField, ForceFieldData, ForceFieldEvaluator
from Scientific.Geometry import Vector
import Numeric, umath

# Class definitions

#
# The base class NonBondedForceField provides the common
# functionality for all non-bonded interactions. The derived
# classes have to deal with determining functional forms
# and parameters and providing the evaluation code
#
class NonBondedForceField(ForceField):

    def __init__(self, name):
	ForceField.__init__(self, name)
	self.type = 'nonbonded'

    def ready(self, global_data):
	return 'bonded' in global_data.get('initialized')

    def excludedPairs(self, subset1, subset2, global_data):
	if 'excluded_pairs' not in global_data.get('initialized'):
	    excluded_pairs = global_data.get('excluded_pairs')
	    if subset1 is not None:
		for s1, s2 in [(subset1, subset2), (subset2, subset1)]:
		    set = {}
		    for a in s1.atomList():
			set[a.index] = None
		    for a in s2.atomList():
			try:
			    del set[a.index]
			except KeyError: pass
		    excluded_pairs = excluded_pairs + \
				     Utility.pairs(set.keys())
		set = {}
		for a in subset1.atomList():
		    set[a.index] = None
		for a in subset2.atomList():
		    set[a.index] = None
		atom_subset = set.keys()
		atom_subset.sort()
	    else:
		atom_subset = None
	    global_data.set('atom_subset', atom_subset)
	    excluded_pairs = map(_normalizePair, excluded_pairs)
	    excluded_pairs.sort(_cmpPair)
            _makeUnique(excluded_pairs)
	    global_data.set('excluded_pairs', excluded_pairs)
	    one_four_pairs = map(_normalizePair,
				 global_data.get('1_4_pairs'))
	    one_four_pairs.sort(_cmpPair)
            _makeUnique(one_four_pairs)
	    global_data.set('1_4_pairs', one_four_pairs)
	    global_data.add('initialized', 'excluded_pairs')
	return global_data.get('excluded_pairs'), \
               global_data.get('1_4_pairs'), \
	       global_data.get('atom_subset')

    def nonbondedList(self, universe, subset1, subset2, global_data):
	try:
	    from MMTK_forcefield import NonbondedList, NonbondedListEvaluator
	except ImportError:
	    return None, None
	nbl = None
	update = None
	if 'nonbondedlist' in global_data.get('initialized'):
	    nbl, cutoff = global_data.get('nonbondedlist')
	if nbl is None:
	    excluded_pairs, one_four_pairs, atom_subset = \
			    self.excludedPairs(subset1, subset2, global_data)
	    excluded_pairs = Numeric.array(excluded_pairs)
	    one_four_pairs = Numeric.array(one_four_pairs)
	    if atom_subset is not None:
		atom_subset = Numeric.array(atom_subset)
	    else:
		atom_subset = Numeric.array([], Numeric.Int)
	    nbl = NonbondedList(excluded_pairs, one_four_pairs, atom_subset,
				universe._spec, self.cutoff)
	    global_data.set('nonbondedlist', (nbl, self.cutoff))
	    global_data.add('initialized', 'nonbondedlist')
	    update = NonbondedListEvaluator(nbl)
	elif (self.cutoff is None and cutoff is not None) or \
	     self.cutoff > cutoff:
	    nbl.setCutoff(self.cutoff)
	return nbl, update

    # the following methods must be overridden by derived classes
    def evaluator(self, universe, subset1, subset2, global_data):
	raise AttributeError

#
# Lennard-Jones force field
#
class LJForceField(NonBondedForceField):

    def __init__(self, name, cutoff):
	NonBondedForceField.__init__(self, name)
	self.cutoff = cutoff

    def evaluator(self, universe, subset1, subset2, global_data):
	n = universe.numberOfPoints()
	lj_type = -Numeric.ones((n,), Numeric.Int)
	atom_types = {}
	for o in universe:
	    for a in o.atomList():
		atom_types[self._atomType(o, a, global_data)] = 1
	i = 0
	for t in atom_types.keys():
	    atom_types[t] = i
	    i = i + 1
	n_types = i
	for o in universe:
	    for a in o.atomList():
		lj_type[a.index] = atom_types[self._atomType(o, a, global_data)]
	eps_sigma = Numeric.zeros((n_types, n_types, 2), Numeric.Float)
	for t, i in atom_types.items():
	    eps_sigma[i,i] = Numeric.array(self._ljParameters(t, global_data))
	eps = eps_sigma[:,:,0]
	sigma = eps_sigma[:,:,1]
	for i in range(n_types):
	    for j in range(i+1, n_types):
		eps[i,j] = umath.sqrt(eps[i,i]*eps[j,j])
		eps[j,i] = eps[i,j]
		sigma[i,j] = 0.5*(sigma[i,i]+sigma[j,j])
		sigma[j,i] = sigma[i,j]
	data = ForceFieldData()
	self.data = data
	data.lj_type = lj_type
	data.lj_14_factor = self.lj_14_factor
	data.universe = universe
	data.eps_sigma = eps_sigma
	data.excluded_pairs, data.one_four_pairs, atom_subset = \
			     self.excludedPairs(subset1, subset2, global_data)
	data.nblist, data.nblist_update = \
		     self.nonbondedList(universe, subset1, subset2, global_data)
	if self.cutoff is None:
	    data.cutoff = 0.
	else:
	    data.cutoff = self.cutoff
	return LJEvaluator(data)

    # the following methods must be overridden by derived classes
    def _atomType(self, o, a, global_data):
	raise AttributeError
    def _ljParameters(self, t, global_data):
	raise AttributeError

def _normalizePair(pair):
    i, j = pair
    if i > j:
	return j, i
    else:
	return i, j

def _cmpPair(p1, p2):
    cmp0 = cmp(p1[0], p2[0])
    if not cmp0:
	return cmp(p1[1], p2[1])
    else:
	return cmp0

def _makeUnique(pair_list):
    i = 0
    while i < len(pair_list):
        item = pair_list[i]
        j = i + 1
        while j < len(pair_list) and pair_list[j] == item:
            del pair_list[j]
        i = i + 1

#
# Electrostatic force field
#
class ElectrostaticForceField(NonBondedForceField):

    def __init__(self, name, cutoff):
	NonBondedForceField.__init__(self, name)
	self.cutoff = cutoff

    def evaluator(self, universe, subset1, subset2, global_data):
	n = universe.numberOfPoints()
	charge = Numeric.zeros((n,), Numeric.Float)
	for o in universe:
	    for a in o.atomList():
		charge[a.index] = self._charge(o, a, global_data)
	data = ForceFieldData()
	self.data = data
	data.charge = charge
	data.es_14_factor = self.es_14_factor
	data.universe = universe
	data.excluded_pairs, data.one_four_pairs, atom_subset = \
			     self.excludedPairs(subset1, subset2, global_data)
	data.nblist, data.nblist_update = \
		     self.nonbondedList(universe, subset1, subset2, global_data)
	if self.cutoff is None:
	    data.cutoff = 0.
	else:
	    data.cutoff = self.cutoff
	return ESEvaluator(data)

    # the following method must be overridden by derived classes
    def _charge(self, o, a, global_data):
	raise AttributeError

#
# Ewald evaluator for electrostatic interactions
#
class ESEwaldForceField(NonBondedForceField):

    def __init__(self, name, options = {}):
	NonBondedForceField.__init__(self, name)
	self.cutoff = options.get('real_cutoff', None)
        self.options = options
        for key in options.keys():
            if key not in self.known_options:
                raise ValueError, key + " is not a recognized option"

    known_options = ['beta', 'real_cutoff', 'reciprocal_cutoff']

    def evaluator(self, universe, subset1, subset2, global_data):
        if not universe.is_periodic:
            raise ValueError, "Ewald methods accepts only periodic universes"
	n = universe.numberOfPoints()
	charge = Numeric.zeros((n,), Numeric.Float)
	for o in universe:
	    for a in o.atomList():
		charge[a.index] = self._charge(o, a, global_data)
	data = ForceFieldData()
	self.data = data
        data.shape = universe.basisVectors()
	data.charge = charge
	data.es_14_factor = self.es_14_factor
	data.universe = universe
	data.excluded_pairs, data.one_four_pairs, data.atom_subset = \
			     self.excludedPairs(subset1, subset2, global_data)
	data.nblist, data.nblist_update = \
		     self.nonbondedList(universe, subset1, subset2,global_data)
        data.options = {}
        data.options['real_cutoff'] = 0.9
        data.options['beta'] = 4.2
        data.options['reciprocal_cutoff'] = 0.1
        for key, value in self.options.items():
            data.options[key] = value
        lx = universe.boxToRealCoordinates(Vector(1., 0., 0.)).length()
        ly = universe.boxToRealCoordinates(Vector(0., 1., 0.)).length()
        lz = universe.boxToRealCoordinates(Vector(0., 0., 1.)).length()
        data.kmax = Numeric.array([lx,ly,lz])/data.options['reciprocal_cutoff']
        data.kmax = Numeric.ceil(data.kmax).astype(Numeric.Int)
	return ESEwaldEvaluator(data)

    # the following methods must be overridden by derived classes
    def _charge(self, o, a, global_data):
	raise AttributeError

#
# Multipole evaluator for electrostatic interactions
#
class ESMPForceField(NonBondedForceField):

    def __init__(self, name, options = {}):
	NonBondedForceField.__init__(self, name)
	self.cutoff = None
        self.options = options
        for key in options.keys():
            if key not in self.known_options:
                raise ValueError, key + " is not a recognized option"

    known_options = ['spatial_decomposition_levels',
                     'multipole_expansion_terms',
                     'use_fft',
                     'fft_blocking_factor',
                     'macroscopic_expansion_terms',
                     'multipole_acceptance']

    def evaluator(self, universe, subset1, subset2, global_data):
	data = ForceFieldData()
	self.data = data
	data.universe = universe
        if universe.is_periodic:
            try:
                data.shape = universe.basisVectors()
            except AttributeError:
                raise ValueError, "Multipole method implemented only " + \
                      "for orthorhombic universes."
        else:
            data.shape = None
	n = universe.numberOfPoints()
	charge = Numeric.zeros((n,), Numeric.Float)
	atom_types = {}
	for o in universe:
	    for a in o.atomList():
		charge[a.index] = self._charge(o, a, global_data)
	data.charge = charge
	data.es_14_factor = self.es_14_factor
	data.excluded_pairs, data.one_four_pairs, data.atom_subset = \
			     self.excludedPairs(subset1, subset2, global_data)
        data.options = {}
        if n < 10000:
            data.options['spatial_decomposition_levels'] = 4
        elif n < 100000:
            data.options['spatial_decomposition_levels'] = 5
        else:
            data.options['spatial_decomposition_levels'] = 6
        data.options['multipole_expansion_terms'] = 8
        data.options['use_fft'] = 0
        data.options['fft_blocking_factor'] = 4
        data.options['macroscopic_expansion_terms'] = 6
        data.options['multipole_acceptance'] = 0.5
        for key, value in self.options.items():
            data.options[key] = value
	return ESMPEvaluator(data)

    # the following methods must be overridden by derived classes
    def _charge(self, o, a, global_data):
	raise AttributeError

#
# Optimized combined Lennard-Jones/electrostatic force field
# for infinite cutoff.
#
class LJESForceField(NonBondedForceField):

    def __init__(self, name):
	NonBondedForceField.__init__(self, name)
	self.cutoff = None

    def evaluator(self, universe, subset1, subset2, global_data):
	n = universe.numberOfPoints()
	lj_type = -Numeric.ones((n,), Numeric.Int)
	charge = Numeric.zeros((n,), Numeric.Float)
	atom_types = {}
	for o in universe:
	    for a in o.atomList():
		atom_types[self._atomType(o, a, global_data)] = 1
		charge[a.index] = self._charge(o, a, global_data)
	i = 0
	for t in atom_types.keys():
	    atom_types[t] = i
	    i = i + 1
	n_types = i
	for o in universe:
	    for a in o.atomList():
		lj_type[a.index] = atom_types[self._atomType(o, a, global_data)]
	eps_sigma = Numeric.zeros((n_types, n_types, 2), Numeric.Float)
	for t, i in atom_types.items():
	    eps_sigma[i,i] = Numeric.array(self._ljParameters(t, global_data))
	eps = eps_sigma[:,:,0]
	sigma = eps_sigma[:,:,1]
	for i in range(n_types):
	    for j in range(i+1, n_types):
		eps[i,j] = umath.sqrt(eps[i,i]*eps[j,j])
		eps[j,i] = eps[i,j]
		sigma[i,j] = 0.5*(sigma[i,i]+sigma[j,j])
		sigma[j,i] = sigma[i,j]
	data = ForceFieldData()
	self.data = data
	data.lj_type = lj_type
	data.charge = charge
	data.lj_14_factor = self.lj_14_factor
	data.es_14_factor = self.es_14_factor
	data.universe = universe
	data.eps_sigma = eps_sigma
	data.excluded_pairs, data.one_four_pairs, data.atom_subset = \
			     self.excludedPairs(subset1, subset2, global_data)
	return LJESEvaluator(data)

    # the following methods must be overridden by derived classes
    def _atomType(self, o, a, global_data):
	raise AttributeError
    def _ljParameters(self, t, global_data):
	raise AttributeError
    def _charge(self, o, a, global_data):
	raise AttributeError

#
# Force field evaluator in Python. For efficiency, force field
# evaluators are normally written in C. But a simple Python version
# using automatic derivatives is a great help in debugging the C
# version!
#
from ForceField import addToGradients, addToForceConstants

class NonBondedList:

    def __init__(self, universe, coordinates, cutoff,
		 excluded_pairs, one_four_pairs, atom_subset):
        self.pairs = []
	if atom_subset is None:
	    atom_subset = range(coordinates.shape[0])
        for a1, a2 in Utility.pairs(atom_subset):
	    d = universe.distance(coordinates[a1], coordinates[a2])
	    if cutoff == 0. or d <= cutoff:
		p = _normalizePair((a1, a2))
		if p not in excluded_pairs:
		    self.pairs.append(p + (p in one_four_pairs,))

    def __len__(self):
	return len(self.pairs)

    def __getitem__(self, item):
	return self.pairs[item]


class LJTestEvaluator(ForceFieldEvaluator):

    def __init__(self, data):
	self.data = data
	self.last_energy = 0.

    name = 'Lennard-Jones'

    def __call__(self, coordinates, gradients, force_constants):
	nb_list = NonBondedList(self.data.universe, coordinates,
				self.data.cutoff, self.data.excluded_pairs,
				self.data.one_four_pairs, self.data.atom_subset)
	energy = 0.
	if force_constants is not None:
	    for i, j, one_four in nb_list:
		ti = self.data.lj_type[i]
		tj = self.data.lj_type[j]
		if ti >= 0 and tj >= 0:
		    eps = self.data.eps_sigma[ti, tj, 0]
		    if one_four:
			eps = eps * self.data.lj_14_factor
		    e, g, fc = lj_e_g_fc(Vector(coordinates[i]),
					 Vector(coordinates[j]),
					 eps, self.data.eps_sigma[ti, tj, 1])
		    energy = energy + e
		    addToGradients(gradients, [i,j], g)
		    addToForceConstants(force_constants, [i,j], fc)
	elif gradients is not None:
	    for i, j, one_four in nb_list:
		ti = self.data.lj_type[i]
		tj = self.data.lj_type[j]
		if ti >= 0 and tj >= 0:
		    eps = self.data.eps_sigma[ti, tj, 0]
		    if one_four:
			eps = eps * self.data.lj_14_factor
		    e, g = lj_e_g(Vector(coordinates[i]),
				  Vector(coordinates[j]),
				  eps, self.data.eps_sigma[ti, tj, 1])
		    energy = energy + e
		    addToGradients(gradients, [i,j], g)
	else:
	    for i, j, one_four in nb_list:
		ti = self.data.lj_type[i]
		tj = self.data.lj_type[j]
		if ti >= 0 and tj >= 0:
		    eps = self.data.eps_sigma[ti, tj, 0]
		    if one_four:
			eps = eps * self.data.lj_14_factor
		    energy = energy + \
			     lj_energy(Vector(coordinates[i]),
				       Vector(coordinates[j]),
				       eps, self.data.eps_sigma[ti, tj, 1])
	self.last_energy = energy
	return energy


class ESTestEvaluator(ForceFieldEvaluator):

    def __init__(self, data):
	self.data = data
	self.last_energy = 0.

    name = 'electrostatic'

    def __call__(self, coordinates, gradients, force_constants):
	nb_list = NonBondedList(self.data.universe, coordinates,
				self.data.cutoff, self.data.excluded_pairs,
				self.data.one_four_pairs, atom_subset)
	energy = 0.
	if force_constants is not None:
	    for i, j, one_four in nb_list:
		qiqj = self.data.charge[i]*self.data.charge[j]
		if one_four:
		    qiqj = qiqj * self.data.es_14_factor
		e, g, fc = es_e_g_fc(Vector(coordinates[i]),
				     Vector(coordinates[j]),
				     qiqj)
		energy = energy + e
		addToGradients(gradients, [i,j], g)
		addToForceConstants(force_constants, [i,j], fc)
	elif gradients is not None:
	    for i, j, one_four in nb_list:
		qiqj = self.data.charge[i]*self.data.charge[j]
		if one_four:
		    qiqj = qiqj * self.data.es_14_factor
		e, g = es_e_g(Vector(coordinates[i]),
			      Vector(coordinates[j]),
			      qiqj)
		energy = energy + e
		addToGradients(gradients, [i,j], g)
	else:
	    for i, j, one_four in nb_list:
		qiqj = self.data.charge[i]*self.data.charge[j]
		if one_four:
		    qiqj = qiqj * self.data.es_14_factor
		energy = energy + es_energy(Vector(coordinates[i]),
					    Vector(coordinates[j]),
					    qiqj)
	self.last_energy = energy
	return energy

#
# Potential energy terms.
#

if 0:
    
    from Potential import PotentialWithGradients, \
         PotentialWithGradientsAndForceConstants

    def lj_energy(r1, r2, eps, sigma):
        dr = r2-r1
        sr6 = (sigma*sigma/(dr*dr))**3
        return 4.*eps*sr6*(sr6-1.)

    lj_e_g = PotentialWithGradients(lj_energy)
    lj_e_g_fc = PotentialWithGradientsAndForceConstants(lj_energy)


    def es_energy(r1, r2, q1q2):
        dr = r2-r1
        return Units.electrostatic_energy*q1q2/dr.length()

    es_e_g = PotentialWithGradients(es_energy)
    es_e_g_fc = PotentialWithGradientsAndForceConstants(es_energy)

#
# Interface to the fast force field evaluator (written in C).
#

try:

    from MMTK_forcefield import LennardJonesEvaluator, CombinedEvaluator

    def LJEvaluator(data):
	nb = data.nblist
	update = data.nblist_update
	ev = LennardJonesEvaluator(data.universe._spec, nb, data.eps_sigma,
				   data.lj_type, data.cutoff,
				   data.lj_14_factor)
	if update is None:
	    return ev
	else:
	    return CombinedEvaluator(Numeric.array([update, ev]))

except ImportError:

    LJEvaluator = LJTestEvaluator


try:

    from MMTK_forcefield import ElectrostaticEvaluator, CombinedEvaluator

    def ESEvaluator(data):
	nb = data.nblist
	update = data.nblist_update
	ev = ElectrostaticEvaluator(data.universe._spec, nb, data.charge,
				    data.cutoff, data.es_14_factor)
	if update is None:
	    return ev
	else:
	    return CombinedEvaluator(Numeric.array([update, ev]))

except ImportError:

    ESEvaluator = ESTestEvaluator


try:

    from MMTK_forcefield import EsEwaldEvaluator, CombinedEvaluator

    def ESEwaldEvaluator(data):
	if data.atom_subset is not None:
            raise ValueError, "Ewald summation not available for subsets"
	nb = data.nblist
	update = data.nblist_update
        if data.universe.basisVectors() is None:
            raise ValueError, "Ewald evaluator needs periodic universe"
        shape = Numeric.array(data.shape, Numeric.Float)
        kcutoff = (2.*Numeric.pi/data.options['reciprocal_cutoff'])**2
        ev = EsEwaldEvaluator(data.universe._spec, shape, nb, data.charge,
                              data.options['real_cutoff'], kcutoff,
                              data.kmax,
                              data.es_14_factor, data.options['beta'])
	if update is None:
	    return ev
	else:
	    return CombinedEvaluator(Numeric.array([update, ev]))

except ImportError:

    pass


try:

    from MMTK_forcefield import EsMPEvaluator

    def ESMPEvaluator(data):
	for pair in data.excluded_pairs:
	    try:
		data.one_four_pairs.remove(pair)
	    except ValueError: pass
	i = 0
	while i < len(data.one_four_pairs)-1:
	    if data.one_four_pairs[i] == data.one_four_pairs[i+1]:
		del data.one_four_pairs[i+1:i+2]
	    else:
		i = i + 1
	excluded_pairs = Numeric.array(data.excluded_pairs)
	one_four_pairs = Numeric.array(data.one_four_pairs)
	if data.atom_subset is None:
	    atom_subset = Numeric.array([], Numeric.Int)
	else:
	    atom_subset = Numeric.array(data.atom_subset)
	nbinfo = [excluded_pairs, one_four_pairs, atom_subset]
        if data.shape is None:
            shape = Numeric.zeros((), Numeric.Float)
        else:
            shape = Numeric.array(data.shape, Numeric.Float)
	ev = EsMPEvaluator(data.universe._spec, shape, nbinfo,
                           data.charge, data.es_14_factor,
                           data.options['spatial_decomposition_levels'],
                           data.options['multipole_expansion_terms'],
                           data.options['use_fft'],
                           data.options['fft_blocking_factor'],
                           data.options['macroscopic_expansion_terms'],
                           data.options['multipole_acceptance'])
	return ev

except ImportError:

    pass


try:

    from MMTK_forcefield import LjEsEvaluator

    def LJESEvaluator(data):
	for pair in data.excluded_pairs:
	    try:
		data.one_four_pairs.remove(pair)
	    except ValueError: pass
	i = 0
	while i < len(data.one_four_pairs)-1:
	    if data.one_four_pairs[i] == data.one_four_pairs[i+1]:
		del data.one_four_pairs[i+1:i+2]
	    else:
		i = i + 1
	excluded_pairs = Numeric.array(data.excluded_pairs)
	one_four_pairs = Numeric.array(data.one_four_pairs)
	if data.atom_subset is None:
	    atom_subset = Numeric.array([], Numeric.Int)
	else:
	    atom_subset = Numeric.array(data.atom_subset)
	nbinfo = [excluded_pairs, one_four_pairs, atom_subset]
	ev = LjEsEvaluator(data.universe._spec, nbinfo,
                           data.eps_sigma, data.lj_type, data.charge,
                           data.lj_14_factor, data.es_14_factor)
	return ev

except ImportError:

    pass
