/* Include file for C force field calculations.
 *
 * Written by Konrad Hinsen
 * last revision: 1999-3-23
 */

#ifndef MMTK_FORCEFIELD_H

/* Common include files */

#include "mmtk.h"
#include "mmtk_universe.h"
#include <math.h>

/* Global variables */

extern double electrostatic_energy_factor;

/* Type definitions */

struct ffeval;

typedef int gradient_function(struct ffeval *evaluator,
			      PyObject *data,
			      int i, vector3 gradient);

typedef int fc_function(struct ffeval *evaluator,
			PyObject *data,
			int i, int j, tensor3 fc,
			double r_sq);

typedef double ff_eval_function(PyObject *evaluator, int natoms,
				PyArrayObject *coordinates,
				PyObject *gradients,
				PyObject *force_constants,
				gradient_function *gradient_fn,
				fc_function *fc_fn,
				int small_change);

typedef void energy_retrieval_function(PyObject *, PyObject *);


/* Evaluator object structure */

typedef struct ffeval {
  PyObject_HEAD
  ff_eval_function *eval_func;
  energy_retrieval_function *add_energy;
  char *name;
  PyUniverseSpecObject *universe_spec;
  PyObject *data1;
  PyObject *data2;
  PyObject *data3;
  PyObject *data4;
  void *scratch;
  double param1, param2, param3, param4;
  double energy;
  double virial;
  double out1, out2, out3, out4;
  int n;
  int error;
  int virial_available;
} PyFFEvaluatorObject;

/* Non-bonded list object structure */

typedef struct {
  double r_sq;
  int i1, i2;
  char one_four;
  char check;
} atom_pair;

typedef struct {
  int *atoms;
  int ix, iy, iz;
  int n, i;
} nbbox;

typedef struct {
  PyObject_HEAD
  PyObject *excluded_pairs;
  PyObject *one_four_pairs;
  PyObject *atom_subset;
  PyUniverseSpecObject *universe_spec;
  int box_count[3];
  int *box_number;
  int *box_atoms;
  nbbox *boxes;
  atom_pair *pairs;
  double cutoff;
  int npairs;
  int allocated_pairs;
  int nboxes;
  int remake;
  int countdown;
} PyNonbondedListObject;

/* Sparse force constant matrix object */

struct pair_fc {
  tensor3 fc;
  int i, j;
};

struct pair_descr {
  int diffij;
  int index;
};

struct pair_descr_list {
  struct pair_descr *list;
  int nalloc;
  int nused;
};

typedef struct sparse_fc {
  PyObject_HEAD
  struct pair_fc *data;
  struct pair_descr_list *index;
  int nalloc;
  int nused;
  int natoms;
  fc_function *fc_fn;
  double cutoff_sq;
} PySparseFCObject; 

/*
 * C API
 */

/* Type definitions */
#define PyFFEvaluator_Type_NUM 0
#define PyNonbondedList_Type_NUM 1
#define PySparseFC_Type_NUM 2

/* Create sparce force constant matrix */
#define PySparseFC_New_RET PySparseFCObject *
#define PySparseFC_New_PROTO Py_PROTO((int natoms, int nalloc))
#define PySparseFC_New_NUM 3

/* Zero sparce force constant matrix */
#define PySparseFC_Zero_RET void
#define PySparseFC_Zero_PROTO Py_PROTO((PySparseFCObject *fc))
#define PySparseFC_Zero_NUM 4

/* Find a pair entry in a sparce force constant matrix */
#define PySparseFC_Find_RET double *
#define PySparseFC_Find_PROTO Py_PROTO((PySparseFCObject *fc, int i, int j))
#define PySparseFC_Find_NUM 5

/* Add a pair contribution to a sparce force constant matrix */
#define PySparseFC_AddTerm_RET int
#define PySparseFC_AddTerm_PROTO Py_PROTO((PySparseFCObject *fc, \
					   int i, int j, double *term))
#define PySparseFC_AddTerm_NUM 6

/* Copy data to an array */
#define PySparseFC_CopyToArray_RET void
#define PySparseFC_CopyToArray_PROTO Py_PROTO((PySparseFCObject *fc, \
					       double *data, int lastdim, \
					       int from1, int to1, \
					       int from2, int to2))
#define PySparseFC_CopyToArray_NUM 7

/* Extract data as an array */
#define PySparseFC_AsArray_RET PyObject *
#define PySparseFC_AsArray_PROTO Py_PROTO((PySparseFCObject *fc, \
					   int from1, int to1, \
					   int from2, int to2))
#define PySparseFC_AsArray_NUM 8

/* Multiply with a vector */
#define PySparseFC_VectorMultiply_RET void
#define PySparseFC_VectorMultiply_PROTO Py_PROTO((PySparseFCObject *fc, \
						  double *result, \
						  double *vector, \
						  int from_i, int to_i, \
						  int from_j, int to_j))
#define PySparseFC_VectorMultiply_NUM 9

/* Create force field evaluator */
#define PyFFEvaluator_New_RET PyFFEvaluatorObject *
#define PyFFEvaluator_New_PROTO Py_PROTO((void))
#define PyFFEvaluator_New_NUM 10

/* Scale by weight vector */
#define PySparseFC_Scale_RET void
#define PySparseFC_Scale_PROTO Py_PROTO((PySparseFCObject *fc, \
					 PyArrayObject *factors))
#define PySparseFC_Scale_NUM 11

/* Total number of C API pointers */
#define PyFF_API_pointers 12


#ifdef _FORCEFIELD_MODULE

/* Type object declarations */
extern PyTypeObject PyFFEvaluator_Type;
extern PyTypeObject PyNonbondedList_Type;
extern PyTypeObject PySparseFC_Type;

/* Type check macros */
#define PyFFEvaluator_Check(op) ((op)->ob_type == &PyFFEvaluator_Type)
#define PyNonbondedList_Check(op) ((op)->ob_type == &PyNonbondedList_Type)
#define PySparseFC_Check(op) ((op)->ob_type == &PySparseFC_Type)

/* C API function declarations */
extern PySparseFC_New_RET PySparseFC_New PySparseFC_New_PROTO;
extern PySparseFC_Zero_RET PySparseFC_Zero PySparseFC_Zero_PROTO;
extern PySparseFC_Find_RET PySparseFC_Find PySparseFC_Find_PROTO;
extern PySparseFC_AddTerm_RET PySparseFC_AddTerm PySparseFC_AddTerm_PROTO;
extern PySparseFC_CopyToArray_RET PySparseFC_CopyToArray \
       PySparseFC_CopyToArray_PROTO;
extern PySparseFC_AsArray_RET PySparseFC_AsArray PySparseFC_AsArray_PROTO;
extern PySparseFC_VectorMultiply_RET PySparseFC_VectorMultiply \
       PySparseFC_VectorMultiply_PROTO;
extern PySparseFC_Scale_RET PySparseFC_Scale PySparseFC_Scale_PROTO;
extern PyFFEvaluator_New_RET PyFFEvaluator_New PyFFEvaluator_New_PROTO;

#else

/* C API address pointer */ 
static void **PyFF_API;

/* Type definitions */
#define PyFFEvaluator_Type *(PyTypeObject *)PyFF_API[PyFFEvaluator_Type_NUM]
#define PyNonbondedList_Type *(PyTypeObject *)PyFF_API[PyNonbondedList_Type_NUM]
#define PySparseFC_Type *(PyTypeObject *)PyFF_API[PySparseFC_Type_NUM]

/* Type check macros */
#define PyFFEvaluator_Check(op) \
   ((op)->ob_type == (PyTypeObject *)PyFF_API[PyFFEvaluator_Type_NUM])
#define PyNonbondedList_Check(op) \
   ((op)->ob_type == (PyTypeObject *)PyFF_API[PyNonbondedList_Type_NUM])
#define PySparseFC_Check(op) \
   ((op)->ob_type == (PyTypeObject *)PyFF_API[PySparseFC_Type_NUM])

/* C API function declarations */
#define PySparseFC_New \
  (*(PySparseFC_New_RET (*)PySparseFC_New_PROTO) \
   PyFF_API[PySparseFC_New_NUM])
#define PySparseFC_Zero \
  (*(PySparseFC_Zero_RET (*)PySparseFC_Zero_PROTO) \
   PyFF_API[PySparseFC_Zero_NUM])
#define PySparseFC_Find \
  (*(PySparseFC_Find_RET (*)PySparseFC_Find_PROTO) \
   PyFF_API[PySparseFC_Find_NUM])
#define PySparseFC_AddTerm \
  (*(PySparseFC_AddTerm_RET (*)PySparseFC_AddTerm_PROTO) \
   PyFF_API[PySparseFC_AddTerm_NUM])
#define PySparseFC_CopyToArray \
  (*(PySparseFC_CopyToArray_RET (*)PySparseFC_CopyToArray_PROTO) \
   PyFF_API[PySparseFC_CopyToArray_NUM])
#define PySparseFC_AsArray \
  (*(PySparseFC_AsArray_RET (*)PySparseFC_AsArray_PROTO) \
   PyFF_API[PySparseFC_AsArray_NUM])
#define PySparseFC_VectorMultiply \
  (*(PySparseFC_VectorMultiply_RET (*)PySparseFC_VectorMultiply_PROTO) \
   PyFF_API[PySparseFC_VectorMultiply_NUM])
#define PyFFEvaluator_New \
  (*(PyFFEvaluator_New_RET (*)PyFFEvaluator_New_PROTO) \
   PyFF_API[PyFFEvaluator_New_NUM])
#define PySparseFC_Scale \
  (*(PySparseFC_Scale_RET (*)PySparseFC_Scale_PROTO) \
   PyFF_API[PySparseFC_Scale_NUM])

#endif

#define MMTK_FORCEFIELD_H
#endif
