# This module implements energy minimizers.
#
# Written by Konrad Hinsen
# last revision: 1999-3-18
#

import Features, Trajectory, Units
from MMTK_minimization import conjugateGradient, steepestDescent
from MMTK_forcefield import TotalEvaluator
import Numeric, operator

#
# Minimizer base class
#
class Minimizer(Trajectory.TrajectoryGenerator):

    def __init__(self, universe, options):
	Trajectory.TrajectoryGenerator.__init__(self, universe, options)

    default_options = {'steps': 100, 'step_size': 0.02*Units.Ang,
		       'convergence': 0.01*Units.kJ/(Units.mol*Units.nm)}

    available_data = ['energy', 'configuration', 'gradients']

    restart_data = ['configuration', 'energy']

    def __call__(self, options):
	raise AttributeError

#
# Steepest descent minimizer
#
class SteepestDescentMinimizer(Minimizer):

    def __init__(self, universe, **options):
	Minimizer.__init__(self, universe, options)
	self.features = [Features.FixedParticleFeature,
			 Features.NoseThermostatFeature,
                         Features.AndersenBarostatFeature]

    def __call__(self, **options):
	self.setCallOptions(options)
	Features.checkFeatures(self, self.universe)
	configuration = self.universe.configuration()
	fixed = self.universe.getAtomBooleanArray('fixed')
	evaluator = self.universe.energyEvaluator().evaluator.CEvaluator()
	evaluator = TotalEvaluator(Numeric.array([evaluator]))
	steepestDescent(self.universe,
			configuration.array, fixed.array, evaluator,
			self.getOption('steps'), self.getOption('step_size'),
			self.getOption('convergence'),
                        self.getActions(),
			'Steepest descent minimization with ' +
			self.optionString(['convergence', 'step_size',
					   'steps']))


#
# Conjugate gradient minimizer
#
class ConjugateGradientMinimizer(Minimizer):

    def __init__(self, universe, **options):
	Minimizer.__init__(self, universe, options)
	self.features = [Features.FixedParticleFeature,
			 Features.NoseThermostatFeature]

    def __call__(self, **options):
	self.setCallOptions(options)
	Features.checkFeatures(self, self.universe)
	configuration = self.universe.configuration()
	fixed = self.universe.getAtomBooleanArray('fixed')
	evaluator = self.universe.energyEvaluator().evaluator.CEvaluator()
	evaluator = TotalEvaluator(Numeric.array([evaluator]))
	conjugateGradient(self.universe,
			  configuration.array, fixed.array, evaluator,
			  self.getOption('steps'), self.getOption('step_size'),
			  self.getOption('convergence'),
                          self.getActions(),
			  'Conjugate gradient minimization with ' +
			  self.optionString(['convergence', 'step_size',
					     'steps']))
