# This module deals with input and output of configurations in PDB format.
#
# Written by Konrad Hinsen
# last revision: 1999-3-19
#

import ChemicalObjects, Collection, Database, Units, Universe, Utility
from Scientific.Geometry import Vector
import Scientific.IO.PDB
import os, string

#
# The chain classes from Scientific.IO.PDB are extended by methods
# that construct MMTK objects and set configurations.
#
class PDBChain:

    def applyConfiguration(self, chain,
                           map = 'pdbmap', alt = 'pdb_alternative'):
        if len(chain) != len(self):
            raise ValueError, "chain lengths do not match"
        for i in range(len(chain)):
            residue = chain[i]
            pdbmap = getattr(residue, map)
            try: altmap = getattr(residue, alt)
            except AttributeError: altmap = {}
            setResidueConfiguration(residue, self[i], pdbmap[0], altmap)
    
class PDBPeptideChain(Scientific.IO.PDB.PeptideChain, PDBChain):

    def createPeptideChain(self, model = 'all',
                           n_terminus=None, c_terminus=None):
        import Proteins
        properties = {'model': model}
        if self.segment_id != '':
            properties['name'] = self.segment_id
        if c_terminus is None:
            properties['c_terminus'] = not self.isTerminated()
        else:
            properties['c_terminus'] = c_terminus
        if n_terminus is not None:
            properties['n_terminus'] = n_terminus
        chain = apply(Proteins.PeptideChain, (self,), properties)
        chain.findHydrogenPositions()
        return chain

class PDBNucleotideChain(Scientific.IO.PDB.NucleotideChain, PDBChain):

    def createNucleotideChain(self, model='all'):
        import NucleicAcids
        properties = {'model': model}
        if self.segment_id != '':
            properties['name'] = self.segment_id
        if self[0].hasPhosphate():
            properties['terminus_5'] = 0
        chain = apply(NucleicAcids.NucleotideChain, (self,), properties)
        chain.findHydrogenPositions()
        return chain

class PDBMolecule(Scientific.IO.PDB.Molecule):

    def createMolecule(self, name=None):
        if name is None:
            name = molecule_names[self.name]
        m = ChemicalObjects.Molecule(name)
        setConfiguration(m, [self])
        return m

#
# The structure class gets some additional methods as well:
#
class PDBConfiguration(Scientific.IO.PDB.Structure):

    def __init__(self, filename):
        filename = Database.PDBPath(filename)
        Scientific.IO.PDB.Structure.__init__(self, filename)
        n = 1
        for residue in self.residues:
            for atom in residue:
                atom.number = n
                n = n + 1

    peptide_chain_constructor = PDBPeptideChain
    nucleotide_chain_constructor = PDBNucleotideChain
    molecule_constructor = PDBMolecule

    def createPeptideChains(self, model='all'):
        import Proteins
        chains = []
        for chain in self.peptide_chains:
            properties = {'model': model}
            if chain.segment_id != '':
                properties['name'] = chain.segment_id
            if not chain.isTerminated():
                properties['c_terminus'] = 0
            chain = apply(Proteins.PeptideChain, (chain,), properties)
            chain.findHydrogenPositions()
            chains.append(chain)
        return chains

    def createNucleotideChains(self, model='all'):
        import NucleicAcids
        chains = []
        for chain in self.nucleotide_chains:
            properties = {'model': model}
            if chain.segment_id != '':
                properties['name'] = chain.segment_id
            if chain[0].hasPhosphate():
                properties['terminus_5'] = 0
            chain = apply(NucleicAcids.NucleotideChain, (chain,), properties)
            chain.findHydrogenPositions()
            chains.append(chain)
        return chains

    def createMolecules(self, names = None, permit_undefined=0):
        collection = Collection.Collection()
        if type(names) == type({}):
            mol_dict = names
            names = None
        else:
            mol_dict = molecule_names
        for name in self.molecules.keys():
            try:
                full_name = mol_dict[name]
            except KeyError:
                full_name = None
            if names is None or name in names or full_name in names:
                if full_name is None and not permit_undefined:
                    raise ValueError, "no definition for molecule " + name
                for molecule in self.molecules[name]:
                    if full_name:
                        m = ChemicalObjects.Molecule(full_name)
                        setConfiguration(m, [molecule])
                    else:
                        m = Collection.Collection()
                        for atom in molecule:
                            try:
                                element = atom['element']
                                a = ChemicalObjects.Atom(element)
                            except KeyError:
                                try:
                                    a = ChemicalObjects.Atom(atom.name[:2])
                                except IOError:
                                    a = ChemicalObjects.Atom(atom.name[:1])
                            a.setPosition(atom.position*Units.Ang)
                            m.addObject(a)
                    collection.addObject(m)
        return collection

    def createAll(self, molecule_names = None, permit_undefined=0):
        collection = Collection.Collection()
        peptide_chains = self.createPeptideChains()
        if peptide_chains:
            import Proteins
            collection.addObject(Proteins.Protein(peptide_chains))
        nucleotide_chains = self.createNucleotideChains()
        collection.addObject(nucleotide_chains)
        molecules = self.createMolecules(None, molecule_names,
                                         permit_undefined)
        collection.addObject(molecules)
        return collection

#
# The following class is a variation for use in Database files.
#
class PDBFile(PDBConfiguration):

    def applyTo(self, object):
        setConfiguration(object, self.residues)

#
# Set atom coordinates from PDB configuration.
#
def setResidueConfiguration(object, pdb_residue, pdbmap, altmap):
    defined = 0
    for atom in pdb_residue:
        name = atom.name
        try: name = altmap[name]
        except KeyError: pass
        try:
            pdbname = pdbmap[1][name]
        except KeyError:
            pdbname = None
            if not object.isSubsetModel():
                raise IOError, 'Atom '+atom.name+' of PDB residue ' + \
                      pdb_residue.name+' not found in residue '+pdbmap[0] + \
                      ' of object ' + object.fullName()
        if pdbname:
            object.setPosition(pdbname, atom.position*Units.Ang)
            object.setIndex(pdbname, atom.number-1)
            defined = defined + 1
    return defined

def setConfiguration(object, pdb_residues,
                     map = 'pdbmap', alt = 'pdb_alternative',
                     toplevel = 1):
    defined = 0
    if hasattr(object, map):
        pdbmap = getattr(object, map)
        try: altmap = getattr(object, alt)
        except AttributeError: altmap = {}
        nres = len(pdb_residues)
        if len(pdbmap) != nres:
            raise IOError, 'PDB configuration does not match object ' + \
                  object.fullName()
        for i in range(nres):
            defined = defined + setResidueConfiguration(object,
                                                        pdb_residues[i],
                                                        pdbmap[i], altmap)
    elif Collection.isCollection(object):
        nres = len(pdb_residues)
        if len(object) != nres:
            raise IOError, 'PDB configuration does not match object ' + \
                  object.fullName()
        for i in range(nres):
            defined = defined + setConfiguration(object[i], [pdb_residues[i]],
                                                 map, alt, 0)
    if toplevel and defined < object.numberOfAtoms():
        name = '[unnamed object]'
        try:
            name = object.fullName()
        except: pass
        if name: name = ' in ' + name
        Utility.warning(`object.numberOfAtoms()-defined` + ' atom(s)' + name +
                        ' were not assigned (new) positions.')
    return defined


#
# Create objects from a PDB configuration.
#
molecule_names = {'HOH': 'water', 'TIP': 'water', 'WAT': 'water'}

def defineMolecule(code, name):
    if molecule_names.has_key(code):
        raise ValueError, "PDB code " + code + " already used"
    molecule_names[code] = name

#
# This object represents a PDB file for output.
#
class PDBOutputFile:

    def __init__(self, filename, subformat= None):
	self.file = Scientific.IO.PDB.PDBFile(filename, 'w', subformat)
	self.warning = 0
	self.atom_no = {}
	self.atom_sequence = []

    def write(self, object, configuration = None, tag = None):
	if not ChemicalObjects.isChemicalObject(object):
	    for o in object:
		self.write(o, configuration)
	else:
	    toplevel = tag is None
	    if toplevel:
		tag = Utility.uniqueAttribute()
	    if hasattr(object, 'pdbmap'):
		for residue in object.pdbmap:
		    self.file.nextResidue(residue[0], )
		    for atom_name, atom in residue[1].items():
			atom = object.getAtom(atom)
			p = atom.position(configuration)
			if Utility.isDefinedPosition(p):
			    try: occ = atom.occupancy
			    except AttributeError: occ = 0.
			    try: temp = atom.temperature_factor
			    except AttributeError: temp = 0.
			    self.file.writeAtom(atom_name, p/Units.Ang,
						occ, temp, atom.type.symbol)
			    self.atom_sequence.append(atom)
			    self.atom_no[id(atom)] = self.file.data['serial_number']
			else:
			    self.warning = 1
			setattr(atom, tag, None)
	    else:
		if hasattr(object, 'is_protein'):
		    for chain in object:		    
			self.write(chain, configuration, tag)
                elif hasattr(object, 'is_chain'):
			self.file.nextChain(None, object.name)
                        for residue in object:
                            self.write(residue, configuration, tag)
                        self.file.terminateChain()
		elif hasattr(object, 'molecules'):
		    for m in object.molecules:
			self.write(m, configuration, tag)
		elif hasattr(object, 'groups'):
		    for g in object.groups:
			self.write(g, configuration, tag)
	    if toplevel:
		for a in object.atomList():
		    if not hasattr(a, tag):
			self.write(a, configuration, tag)
		    delattr(a, tag)

    def close(self):
	self.file.close()
	if self.warning:
	    Utility.warning('Some atoms are missing in the output file ' + \
			    'because their positions are undefined.')
	    self.warning = 0

class PDBConnectOutputFile(PDBOutputFile):
    def __init__(self, filename):
	PDBOutputFile.__init__(self, filename)
	self.top_level = 1

    def write(self, object, configuration = None, tag = None):
	PDBOutputFile.write(self, object, configuration, tag)
	if tag is None:
	    if hasattr(object, 'bonds_to_atom'):
		id_list = object.bonds_to_atom.keys()
		id_list.sort()
		for id_a in id_list:
		    blist = object.bonds_to_atom[id_a]
		    connect_list = []
		    double_bonds = []
		    triple_bonds = []
		    for b in blist:
			try:
			    a2 = b.otherAtomFromId(id_a)
			except ValueError:
			    pass # ignore DanglingBond.otherAtomName complaints
			else:
			    a_no = self.atom_no[id(a2)]
			    connect_list.append(a_no)
			    if b.order >= 2:
				double_bonds.append(a_no)
			    if b.order >= 3:
				triple_bonds.append(a_no)
		    try:
			a_no = self.atom_no[id_a]
		    except KeyError: pass
		    else:
			self.writeBond(a_no, connect_list)
			if double_bonds:
			    self.writeBond(a_no, double_bonds)
			if triple_bonds:
			    self.writeBond(a_no, triple_bonds)
	    else:
		Utility.warning('No CONECT lines written because no '
				+ 'bonds_to_atom list found.')

		# writeBond probably belongs in Scientific.IO.PDB
    def writeBond(self, number, connect_list):
	self.file.data['serial_number'] = number
	self.file.data['bonded'] = connect_list
	self.file.writeLine('CONECT', self.file.data)

