# This module handles input and output of PDB files.
# It deals only with the file format, not with the
# structure of any objects in such files beyond
# atoms, residues, and chains.
#
# Written by Konrad Hinsen <hinsen@ibs.ibs.fr>
# with help from Lutz Ehrlich <Lutz.Ehrlich@EMBL-Heidelberg.de>
# Last revision: 1999-2-25
# 

"""This module provides classes that represent PDB files and
configurations contained in PDB files. It provides access
to PDB files on two levels: low-level (line by line) and
high-level (residues and atoms).

Attention: The PDB file format has been heavily abused, and it is
probably impossible to write code that can deal with all variants
correctly. This modules tries to read the widest possible range
of PDB files, but gives priority to correctly interpreting the
PDB format as defined by the Brookhaven National Laboratory.

Class PDBFile:
--------------
Creation:
   PDBFile(filename, mode)
     mode is 'r' for reading and 'w' for writing, default is 'r'
     If the filename ends with '.Z' or '.gz', it is automatically
     compressed or uncompressed using 'compress' or 'gzip'.
     The filename may contain the abbreviations '~' or '~user'
     to refer to home directories.

Low-level access methods:
   readLine()
     returns the next non-blank line. The return
     value is a tuple whose first element contains the line type
     identifier. For lines of type 'ATOM' and 'HETATM', the
     remaining fields are put into a dictionary which is returned
     as the second tuple element. The entry 'position' is an
     object of class Vector, the other entries are strings or numbers.
     For other lines, the second element of the tuple contains the part
     of the line after the type field as a single string.
   writeComment(text)
     writes the supplied text into one or several comment lines.
     Each line of the text is prefixed with 'REMARK' and written
     to the file.
   nextResidue(name)
     must be called in between writeAtom calls to indicate the
     beginning of a new residue. The argument is the residue
     identifier.
   writeAtom(name, position, occupancy=0., temp_factor=0., het_flag=0)
     writes an ATOM or HETATM line for an atom.

High-level access methods:
   readConfiguration()
     reads all ATOM and HETATM lines, ignoring anything else.
     The result is a list of Residue objects (see below).
     If alternate locations are provided for any atom, only
     the first version is used.
     This method is intended for PDB files that do not contain
     proteins; for proteins the method readSequenceWithConfiguration()
     is more useful.
   readSequence(conf = 0)
     returns a list of peptide chains in the file, each peptide
     chain being defined by a list of residues. If conf==0, each
     residue is represented by a three-character string. If conf==1,
     each residue is represented by a Residue object (see below).
     If alternate locations are provided for any atom, only
     the first version is used.
     Note: peptide chains are distinguished from other molecules
     by the use of ATOM records (as opposed to HETATM). The end
     of each chain is determined by looking for residues that
     have one more heavy atom (the second oxygen) than the
     standard versions. Neither of these assumptions works in
     all cases.
   readSequenceWithConfiguration()
     equivalent to readSequence(1)
   writeConfiguration(conf)
     writes the supplied configuration (a list of Residue objects)
     to the file. This method can be called several times with
     parts of a configuration; however, each residue must be
     written in a single call.

General methods:
   close()
     *must* be called after writing to a file, should be called after
     reading to avoid memory leaks.


Class PDBConfiguration:
-----------------------
Creation:
   PDBConfiguration(filename)
     to generate a configuration by reading a file (see PDBFile).
   PDBConfiguration(residue_list)
     to generate a configuration from a given list of Residue
     objects (see below).
   PDBConfiguration()
     to generate an empty configuration.

Access (assuming c is an instance of PDBConfiguration):
   len(c)
     returns the number of Residues in the configuration
   c[i]
     returns residue number i.
   c.writeToFile(filename)
     writes the configuration to a file.
   c.addResidue(residue)
     adds a residue to the configuration.


Class Residue:
--------------
Creation:
   Residue(name, atoms)
     to generate a residue with identifier 'name' and atoms
     from the list 'atoms', whose elements must be instances
     of class Atom (see below).
   Residue(name)
     to generate an empty residue with a given identifier.

Access (assuming r is an instance of Residue):
   len(r)
     returns the number of atoms in the residue
   r[i]
     returns atom number i.
   r.addAtom(atom)
     adds an atom to the residue.

Class Atom:
-----------
Creation:
   Atom(name, position, number = None)
     to generate an atom with identifier 'name' at a given position.
     The position must be an instance of class Vector.
     The optional number is used to store the atom sequence number when
     reading from PDB files.

Modification methods:
   moveBy(distance)
     adds the given distance (instance of class Vector) to the
     position.
   moveTo(position)
     changes the position.

Example:
--------
conf = PDBConfiguration('example.pdb')
print conf
for residue in conf:
    for atom in residue:
        print atom
new_conf = PDBConfiguration(conf)
new_conf.writeToFile('test.pdb')
"""

from Scientific.IO.TextFile import TextFile
from Scientific.Geometry import Vector
import string

class Group:

    def __init__(self, name, atoms = None):
	self.name = name
	self.atom_list = []
	self.atoms = {}
	if atoms:
	    self.atom_list = atoms
	    for a in atoms:
		self.atoms[a.name] = a

    def __len__(self):
	return len(self.atom_list)

    def __getitem__(self, item):
	if type(item) == type(0):
	    return self.atom_list[item]
	else:
	    return self.atoms[item]

    def __str__(self):
	s = 'Residue ' + self.name + ':\n'
	for atom in self.atom_list:
	    s = s + '  ' + `atom` + '\n'
	return s

    __repr__ = __str__

    def addAtom(self, atom):
	self.atom_list.append(atom)
	self.atoms[atom.name] = atom

    def deleteAtom(self, atom):
	self.atom_list.remove(atom)
	del self.atoms[atom.name]

    def deleteHydrogens(self):
        delete = []
        for a in self.atom_list:
            if a.name[0] == 'H' or (a.name[0] in string.digits
                                    and a.name[1] == 'H'):
                delete.append(a)
	for a in delete:
	    self.deleteAtom(a)

    def changeName(self, name):
	self.name = name

class Residue(Group):

    def __init__(self, name, atoms = None, number = None):
	Group.__init__(self, name, atoms)
	self.number = number

    def __str__(self):
	s = 'Residue ' + self.name + ':\n'
	for atom in self.atom_list:
	    s = s + '  ' + `atom` + '\n'
	return s

    __repr__ = __str__


class AminoAcidResidue(Residue):

    def isCTerminus(self):
        return self.atoms.has_key('OXT') or self.atoms.has_key('OT2')


class NucleotideResidue(Residue):

    def __init__(self, name, atoms = None, number = None):
        name = name[:-1] + 'D' + name[-1]
	Residue.__init__(self, name, atoms, number)

    def foundRibose(self):
        self.name = self.name[:-2] + 'R' + self.name[-1]

    def hasRibose(self):
        return self.atoms.has_key('O2*')

    def hasDesoxyribose(self):
        return not self.hasRibose()

    def hasPhosphate(self):
        return self.atoms.has_key('P')


class Molecule(Group):

    def __init__(self, name, atoms = None):
	Group.__init__(self, name, atoms)

    def __str__(self):
	s = 'Molecule ' + self.name + ':\n'
	for atom in self.atom_list:
	    s = s + '  ' + `atom` + '\n'
	return s

    __repr__ = __str__


class Atom:

    def __init__(self, name, position, number = None,
		 occupancy = 0., temp_factor = 0.):
	self.name = name
	self.position = position
	self.occupancy = occupancy
	self.temp_factor = temp_factor
	if number is not None:
	    self.number = number

    def __str__(self):
	return 'Atom ' + self.name + ' at ' + str(self.position)

    __repr__ = __str__

    def moveBy(self, distance):
	self.position = self.position + distance

    def moveTo(self, position):
	self.position = position

    def type(self):
	return 'ATOM  '

class HetAtom(Atom):

    def type(self):
	return 'HETATM'
    
    def __str__(self):
	return 'HetAtom ' + self.name + ' at ' + str(self.position)

    __repr__ = __str__


class PDBFile:

    def __init__(self, filename, mode = 'r'):
	self.file = TextFile(filename, mode)
	self.atom_num = 0
	self.chain_num = -1
	self.chain_id = ' '
	self.abs_res_num = 0
	self.rel_res_num = 0
	self.output = string.lower(mode[0]) == 'w'

    def readLine(self):
	while 1:
	    line = self.file.readline()
	    if not line: return ('END','')
	    if line[-1] == '\n': line = line[:-1]
	    line = string.strip(line)
	    if line: break
	line = (line + 80*' ')[:80]
	type, line = (string.strip(line[:6]), line[6:])
	if type == 'ATOM' or type == 'HETATM':
	    info = {'serial_number': _atoi(line[:5]),
		    'name': string.strip(line[6:10]),
		    'alternate': string.strip(line[10]),
		    'residue_name': string.strip(line[11:14]),
		    'chain_id': string.strip(line[15]),
		    'residue_number': _residueNumber(line[16:21]),
		    'position': Vector(_atof(line[24:32]), _atof(line[32:40]),
				       _atof(line[40:48])),
		    'occupancy': _atof(line[48:54]),
		    'temperature_factor': _atof(line[54:60]),
		    'segment_id': string.strip(line[66:70]),
		    'element': string.strip(line[70:72]),
		    'charge': string.strip(line[72:74])}
	    return (type, info)
	elif type == 'CONECT':
	    bonded = filter(lambda i: i > 0,
			    [_atoi(line[6:10]), _atoi(line[11:15]),
			     _atoi(line[16:20]), _atoi(line[21:25])])
	    hydrogen_bonded = filter(lambda i: i > 0,
				     [_atoi(line[26:30]), _atoi(line[31:35]),
				      _atoi(line[41:45]), _atoi(line[46:50])])
	    salt_bridged = filter(lambda i: i > 0,
				  [_atoi(line[36:40]), _atoi(line[51:55])])
	    info = {'serial_number': _atoi(line[:5]),
		    'bonded': bonded,
		    'hydrogen_bonded': hydrogen_bonded,
		    'salt_bridged': salt_bridged}
	    return (type, info)
	else:
	    return (type, line)

    def readConfiguration(self):
	resnum = -1
	conf = PDBConfiguration()
	while 1:
	    type, info = self.readLine()
	    if type == 'END': break
	    if type == 'ATOM' or type == 'HETATM':
		if info['residue_number'] != resnum:
		    resnum = info['residue_number']
		    residue = Residue(info['residue_name'], None, resnum)
		    conf.addResidue(residue)
		    alternate = None
		take = 1
		alt = info['alternate']
		if alt:
		    if alternate is None:
			alternate = alt
		    take = alternate == alt
		if take:
		    if type=='ATOM':
			a = Atom( info['name'],info['position'],
				  info['serial_number'], info['occupancy'],
				  info['temperature_factor'])
		    elif type=='HETATM':
			a = HetAtom( info['name'], info['position'],
				     info['serial_number'],
				     info['occupancy'],
				     info['temperature_factor'])
		    residue.addAtom(a)

	return conf

    def readSequence(self, conf = 0):
	chains = PDBSequenceConfiguration()
	sequence = None
	chain_id = None
	resnum = -1
	residue = None
        resname = None
	natoms = 0
	while 1:
	    type, info = self.readLine()
	    if type == 'END': break
	    if (type == 'ATOM' or type == 'HETATM'):
		last_residue = resname
		atom = info['name']
		resname = info['residue_name']
		amino = self.heavy_atoms.has_key(resname)
                nucleic = resname in self.nucleic_acids
		water = resname in ['HOH', 'WAT', 'TIP']
		if water:
		    resname = 'HOH'
		new_chain = info['chain_id'] != chain_id
		if info['residue_number'] != resnum \
		   or resname != last_residue or new_chain:
		    resnum = info['residue_number']
		    chain_id = info['chain_id']
		    if amino:
			if new_chain: sequence = None
			else:
			    try:
				if natoms > self.heavy_atoms[last_residue]:
				    sequence = None
			    except KeyError: pass
			if sequence is None:
			    sequence = PDBPeptideChain()
			    chains.addChain(sequence)
			residue = AminoAcidResidue(resname, None, resnum)
			sequence.addResidue(residue)
                    elif nucleic:
			if new_chain: sequence = None
			if sequence is None:
			    sequence = PDBNucleotideChain()
			    chains.addChain(sequence)
			residue = NucleotideResidue(resname, None, resnum)
			sequence.addResidue(residue)
		    elif water:
			residue = Molecule(resname)
			chains.addWater(residue)
			sequence = None
		    else:
			residue = Residue(resname, None, resnum)
			chains.addOther(residue)
			sequence = None
		    natoms = 0
		    alternate = None
		take = 1
		alt = info['alternate']
		if alt:
		    if alternate is None:
			alternate = alt
		    take = alternate == alt
		if take:
		    if conf:
			if water and atom[0] == 'O':
			    atom = 'O'
			if type=='ATOM':
			    a = Atom(atom, info['position'],
				     info['serial_number'],
				     info['occupancy'],
				     info['temperature_factor'])
			elif type=='HETATM':
			    a = HetAtom(atom, info['position'],
					info['serial_number'],
					info['occupancy'],
					info['temperature_factor'])
			residue.addAtom(a)
                    if nucleic and atom == 'O2*':
                        residue.foundRibose()
		    if not (atom[0] == 'H' or
			    (atom[0] in string.digits and atom[1] == 'H')):
			natoms = natoms + 1

	if conf:
	    return chains
	else:
	    return chains.sequence()

    heavy_atoms = {'ALA': 5, 'ARG': 11, 'ASN': 8, 'ASP': 8,
		   'CYS': 6, 'CYX': 6,  'GLN': 9, 'GLU': 9, 'GLY': 4,
		   'HIS': 10, 'HSD': 10, 'HSE': 10, 'HSP': 10,
		   'ILE': 8, 'LEU': 8,
		   'LYS': 9, 'MET': 8, 'PHE': 11, 'PRO': 7,
		   'SER': 6, 'THR': 7, 'TRP': 14, 'TYR': 12, 'VAL': 7,
		   'ACE': 3, 'NME': 1}

    nucleic_acids = ['A', 'C', 'G', 'I', 'T', 'U',
                     '+A', '+C', '+G', '+I', '+T', '+U']

    def readSequenceWithConfiguration(self):
	return self.readSequence(1)

    def writeComment(self, text):
	while text:
	    eol = string.find(text,'\n')
	    if eol == -1:
		eol = len(text)
	    self.file.write('REMARK %s \n' % text[:eol])
	    text = text[eol+1:]

    def nextChain(self):
	self.chain_num = self.chain_num + 1
	self.chain_id = self._chain_ids[self.chain_num % len(self._chain_ids)]
	self.rel_res_num = 0

    def noChain(self):
	self.chain_id = ' '
	self.rel_res_num = 0

    _chain_ids = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'

    def nextResidue(self, resid):
	self.resid = string.upper(resid)
	self.abs_res_num = self.abs_res_num + 1
	self.rel_res_num = self.rel_res_num + 1

    def writeAtom(self, name, position, occupancy=0., temp_factor=0.,
		  het_flag = 0):
	name = string.upper(name)
	self.atom_num = self.atom_num + 1
	if het_flag:
	    format = 'HETATM%5d '
	else:
	    format = 'ATOM  %5d '
	self.file.write(format % (self.atom_num))
        if len(name) < 4 and name[0] not in string.digits:
	    name = ' ' + name
	self.file.write(string.ljust(name,4)[0:4] + ' ')
	self.file.write(string.ljust(self.resid,3)[0:3] + ' ')
	self.file.write(self.chain_id)
	self.file.write('%4d    ' % self.rel_res_num)
	self.file.write('%8.3f' % position[0])
	self.file.write('%8.3f' % position[1])
	self.file.write('%8.3f' % position[2])
	self.file.write('%6.2f' % occupancy)
	self.file.write('%6.2f' % temp_factor)
	self.file.write('\n')

    two_letter_elements = ['HE', 'LI', 'BE', 'NE', 'NA', 'MG', 'AL', 'SI',
			   'CL', 'AR', 'CA', 'SC', 'TI', 'CR', 'MN', 'FE',
			   'CO', 'NI', 'CU', 'ZN', 'GA', 'GE', 'AS', 'SE',
			   'BR', 'KR', 'RB', 'SR', 'ZR', 'NB', 'MO', 'TC',
			   'RU', 'RH', 'PD', 'AG', 'CD', 'IN', 'SN', 'SB',
			   'TE', 'XE', 'CS', 'BA', 'HF', 'TA', 'RE', 'OS',
			   'IR', 'PT', 'AU', 'HG', 'TL', 'PB', 'BI', 'PO',
			   'AT', 'RN', 'FR', 'RA', 'RF', 'HA', 'SG', 'NS',
			   'HS', 'MT', 'LA', 'CE', 'PR', 'ND', 'PM', 'SM',
			   'EU', 'GD', 'TB', 'DY', 'HO', 'ER', 'TM', 'YB',
			   'LU', 'AC', 'TH', 'PA', 'NP', 'PU', 'AM', 'CM',
			   'BK', 'CF', 'ES', 'FM', 'MD', 'NO', 'LR']

    def writeConfiguration(self, data):
	for residue in data:
	    self.nextResidue(residue.name)
	    for atom in residue.atom_list:
		self.writeAtom(atom.name, atom.position, atom.occupancy,
			       atom.temp_factor, atom.type() == 'HETATM')

    def close(self):
	if self.output:
	    self.file.write('END\n')
	self.file.close()


def defineResidue(symbol, number_of_atoms):
    PDBFile.heavy_atoms[string.upper(symbol)] = number_of_atoms

#
# Fortran-compatible conversion functions
#
def _atoi(s):
    s = string.strip(s)
    if s:
	return string.atoi(s)
    else:
	return 0

def _atof(s):
    s = string.strip(s)
    if s:
	return string.atof(s)
    else:
	return 0.

#
# Residue number parser, can handle insertion codes
#
def _residueNumber(s):
    if s[-1] in string.digits:
        return _atoi(s)
    else:
        return (_atoi(s[:-1]), s[-1])

#
# Configuration class
#
class PDBConfiguration:

    def __init__(self, conf = None):
	if conf is None:
	    self.residues = []
	elif type(conf) == type(''):
	    file = PDBFile(conf)
	    self.residues = file.readConfiguration().residues
	    file.close()
	else:
	    self.residues = conf

    def __len__(self):
	return len(self.residues)

    def __getitem__(self, item):
	return self.residues[item]

    def __getslice__(self, i1, i2):
	return PDBConfiguration(self.residues[i1:i2])

    def __delslice__(self, i1, i2):
	del self.residues[i1:i2]

    def __str__(self):
	s = ''
	for residue in self.residues:
	    s = s + str(residue)
	return s
    __repr__ = __str__

    def addResidue(self, residue):
	self.residues.append(residue)

    def deleteHydrogens(self):
	for r in self.residues:
	    r.deleteHydrogens()

    def hasChains(self):
	return 0

    def writeToFile(self, filename):
	file = PDBFile(filename, 'w')
	file.writeConfiguration(self.residues)
	file.close()

class PDBPeptideChain(PDBConfiguration):

    def sequence(self):
	return map(lambda r: r.name, self.residues)

    def __getslice__(self, i1, i2):
	return PDBPeptideChain(self.residues[i1:i2])

class PDBNucleotideChain(PDBConfiguration):

    def sequence(self):
	return map(lambda r: r.name, self.residues)

    def __getslice__(self, i1, i2):
	return PDBNucleotideChain(self.residues[i1:i2])

class PDBSequenceConfiguration:

    def __init__(self, chains = None, water = None, other = None):
	if chains is None:
	    self.chains = []
	else:
	    self.chains = chains
	if water is None:
	    self.water = []
	else:
	    self.water = water
	if other is None:
	    self.other = []
	else:
	    self.other = other

    def __len__(self):
	return len(self.chains)

    def __getitem__(self, item):
	return self.chains[item]

    def __getslice__(self, i1, i2):
	return PDBSequenceConfiguration(self.chains[i1:i2])

    def __setitem__(self, item, value):
        self.chains[item] = value

    def hasChains(self):
	return 1

    def addChain(self, chain):
	self.chains.append(chain)

    def addWater(self, molecule):
	self.water.append(molecule)

    def addOther(self, residue):
	self.other.append(residue)

    def splitChain(self, chain, position):
	c = self.chains[chain]
	c1 = c[:position]
	c2 = c[position:]
	self.chains[chain] = c2
	self.chains.insert(chain, c1)

    def peptideChains(self):
	return self.chains

    def waterMolecules(self):
	return PDBConfiguration(self.water)

    def unknownResidues(self):
	return PDBConfiguration(self.other)

    def sequence(self):
	return map(lambda c: c.sequence(), self.chains)

    def deleteHydrogens(self):
	for c in self.chains:
	    c.deleteHydrogens()
	for m in self.water:
	    m.deleteHydrogens()

    def writeToFile(self, filename):
	file = PDBFile(filename, 'w')
	for chain in self.chains:
	    file.writeConfiguration(chain.residues)
	if self.water:
	    file.writeConfiguration(self.water)
	file.close()

if __name__ == '__main__':

    print 'sequence'
    s = PDBFile('165d.pdb').readSequence()
    print s
    print 'configuration'
    sc = PDBFile('165d.pdb').readSequence(1)
    print map(len, sc)
