/* Deformation energy calculation.
 *
 * Written by Konrad Hinsen
 * last revision: 1999-3-10
 */

#include "mmtk_universe.h"
#include "mmtk_forcefield.h"
#include "mmtk_forcefield_private.h"

/*
 * Deformation evaluation
 */
static double
deformation(vector3 *x, vector3 *v, vector3 *g, double *la, int natoms,
	    PyNonbondedListObject *nblist,
	    double fc_length, double factor, int normalize)
{
  int n, npairs;
  double cutoff_sq;
  double norm, def;
  atom_pair *pair;

  if (normalize) {
    norm = 0.;
    for (n = 0; n < natoms; n++)
      norm += v[n][0]*v[n][0] + v[n][1]*v[n][1] + v[n][2]*v[n][2];
    norm = sqrt(norm/natoms);
  }
  else
    norm = 1.;

  if (g != NULL)
    for (n = 0; n < natoms; n++)
      g[n][0] = g[n][1] = g[n][2] = 0.;

  if (la != NULL)
    for (n = 0; n < natoms; n++)
      la[n]= 0.;

  def = 0.;
  npairs = 0;
  cutoff_sq = nblist->cutoff*nblist->cutoff;
  pair = nblist->pairs;
  for (n = 0; n < nblist->npairs; n++, pair++)
    if (pair->r_sq <= cutoff_sq) {
      int i = pair->i1;
      int j = pair->i2;
      vector3 rij, vij;
      double rij2, k, l, l2;
      rij[0] = x[i][0]-x[j][0];
      rij[1] = x[i][1]-x[j][1];
      rij[2] = x[i][2]-x[j][2];
      vij[0] = v[i][0]-v[j][0];
      vij[1] = v[i][1]-v[j][1];
      vij[2] = v[i][2]-v[j][2];
      rij2 = dot(rij, rij);
      k = factor * exp((0.01-rij2)/sqr(fc_length));
      l = dot(rij, vij)/norm;
      l2 = k*l*l/rij2;
      if (g != NULL) {
	g[i][0] += 2*k*l*rij[0]/(norm*natoms*rij2);
	g[i][1] += 2*k*l*rij[1]/(norm*natoms*rij2);
	g[i][2] += 2*k*l*rij[2]/(norm*natoms*rij2);
	g[j][0] -= 2*k*l*rij[0]/(norm*natoms*rij2);
	g[j][1] -= 2*k*l*rij[1]/(norm*natoms*rij2);
	g[j][2] -= 2*k*l*rij[2]/(norm*natoms*rij2);
      }
      if (la != NULL) {
	la[i] += 0.5*l2;
	la[j] += 0.5*l2;
      }
      def += l2;
      npairs++;
    }
  if (g != NULL) {
    if (normalize)
      for (n = 0; n < natoms; n++) {
	g[n][0] = g[n][0] - 2.*def*v[n][0]/(sqr(natoms)*sqr(norm));
	g[n][1] = g[n][1] - 2.*def*v[n][1]/(sqr(natoms)*sqr(norm));
	g[n][2] = g[n][2] - 2.*def*v[n][2]/(sqr(natoms)*sqr(norm));
      }
  }
  return def/natoms;
}

static double
finite_deformation(vector3 *x, vector3 *v, vector3 *g, double *la, int natoms,
		   PyNonbondedListObject *nblist,
		   double fc_length, double factor)
{
  int n, npairs;
  double cutoff_sq;
  double def;
  atom_pair *pair;

  if (g != NULL)
    for (n = 0; n < natoms; n++)
      g[n][0] = g[n][1] = g[n][2] = 0.;

  if (la != NULL)
    for (n = 0; n < natoms; n++)
      la[n]= 0.;

  def = 0.;
  npairs = 0;
  cutoff_sq = nblist->cutoff*nblist->cutoff;
  pair = nblist->pairs;
  for (n = 0; n < nblist->npairs; n++, pair++)
    if (pair->r_sq <= cutoff_sq) {
      int i = pair->i1;
      int j = pair->i2;
      vector3 rij, rvij;
      double rij2, rvij2, rrvij2, k, l, l2;
      rij[0] = x[i][0]-x[j][0];
      rij[1] = x[i][1]-x[j][1];
      rij[2] = x[i][2]-x[j][2];
      rvij[0] = rij[0] + v[i][0]-v[j][0];
      rvij[1] = rij[1] + v[i][1]-v[j][1];
      rvij[2] = rij[2] + v[i][2]-v[j][2];
      rij2 = dot(rij, rij);
      k = factor * exp((0.01-rij2)/sqr(fc_length));
      rvij2 = dot(rvij, rvij);
      rrvij2 = sqrt(rvij2);
      l = rrvij2-sqrt(rij2);
      l2 = k*l*l;
      if (g != NULL) {
	double f = 2.*k*l/(natoms*rrvij2);
	g[i][0] += f*rvij[0];
	g[i][1] += f*rvij[1];
	g[i][2] += f*rvij[2];
	g[j][0] -= f*rvij[0];
	g[j][1] -= f*rvij[1];
	g[j][2] -= f*rvij[2];
      }
      if (la != NULL) {
	la[i] += 0.5*l2;
	la[j] += 0.5*l2;
      }
      def += l2;
      npairs++;
    }
  return def/natoms;
}

static PyObject *
deformation_py(PyObject *dummy, PyObject *args)
{
  PyArrayObject *configuration;
  PyArrayObject *displacement;
  PyArrayObject *gradient = NULL;
  PyArrayObject *l_atom = NULL;
  PyNonbondedListObject *nblist;

  int natoms;
  vector3 *x, *v, *g;
  double *la;
  double fc_length = 0.3;
  double factor = 1.;
  int normalize = 0, finite = 0;

  if (!PyArg_ParseTuple(args, "O!O!O!OOddi|i",
			&PyArray_Type, &configuration,
			&PyArray_Type, &displacement,
			&PyNonbondedList_Type, &nblist,
			&gradient, &l_atom,
			&fc_length, &factor, &normalize, &finite))
    return NULL;
  natoms = configuration->dimensions[0];
  x = (vector3 *)configuration->data;
  v = (vector3 *)displacement->data;
  if ((PyObject *)gradient == Py_None)
    g = NULL;
  else {
    if (PyArray_Check(gradient))
      g = (vector3 *)gradient->data;
    else {
      PyErr_SetString(PyExc_TypeError, "not an array");
      return NULL;
    }
  }
  if ((PyObject *)l_atom == Py_None)
    la = NULL;
  else {
    if (PyArray_Check(l_atom))
      la = (double *)l_atom->data;
    else {
      PyErr_SetString(PyExc_TypeError, "not an array");
      return NULL;
    }
  }
  if (finite)
    return PyFloat_FromDouble(finite_deformation(x, v, g, la, natoms, nblist,
						 fc_length, factor));
  else
    return PyFloat_FromDouble(deformation(x, v, g, la, natoms, nblist,
					  fc_length, factor, normalize));
}

/* Deformation reduction */

void
reduce_deformation(vector3 *x, vector3 *v, vector3 *g, int natoms, int niter,
		   PyNonbondedListObject *nblist,
		   double fc_length, double factor)
{
  double f, min_dist_sq, max_k;
  int i, j;

  min_dist_sq = 1.e30;
  for (i = 0; i < nblist->npairs; i++) {
    double r_sq = nblist->pairs[i].r_sq;
    if (r_sq < min_dist_sq)
      min_dist_sq = r_sq;
  }
  max_k = factor * exp((0.01-min_dist_sq)/sqr(fc_length));
  f = 0.9/max_k;

  for (i = 0; i < niter; i++) {
    deformation(x, v, g, NULL, natoms, nblist, fc_length, factor, 0);
    for (j = 0; j < natoms; j++) {
      v[j][0] -= f*g[j][0];
      v[j][1] -= f*g[j][1];
      v[j][2] -= f*g[j][2];
    }
  }
}

void
reduce_finite_deformation(vector3 *x, vector3 *v, vector3 *g,
			  int natoms, double rms_reduction,
			  PyNonbondedListObject *nblist,
			  double fc_length, double factor)
{
  double rms_sq, rms_limit, rms_sq_last, rms_grad, scale, step;
  int i;

  rms_sq = 0.;
  for (i = 0; i < natoms; i++)
    rms_sq += v[i][0]*v[i][0] + v[i][1]*v[i][1] + v[i][2]*v[i][2];
  rms_limit = sqrt(rms_sq/natoms)-rms_reduction;
  if (rms_limit < 0.)
    rms_limit = 0.;
  rms_limit = natoms*rms_limit*rms_limit;
  step = 0.01;
  while (1) {
    if (rms_sq <= rms_limit)
      break;
    finite_deformation(x, v, g, NULL, natoms, nblist, fc_length, factor);
    rms_grad = 0.;
    for (i = 0; i < natoms; i++)
      rms_grad += g[i][0]*g[i][0] + g[i][1]*g[i][1] + g[i][2]*g[i][2];
    rms_sq_last = rms_sq;
    while (1) {
      scale = step/sqrt(rms_grad);
      for (i = 0; i < natoms; i++) {
	v[i][0] -= scale*g[i][0];
	v[i][1] -= scale*g[i][1];
	v[i][2] -= scale*g[i][2];
      }
      rms_sq = 0.;
      for (i = 0; i < natoms; i++)
	rms_sq += v[i][0]*v[i][0] + v[i][1]*v[i][1] + v[i][2]*v[i][2];
      if (rms_sq > rms_sq_last) {
	for (i = 0; i < natoms; i++) {
	  v[i][0] += scale*g[i][0];
	  v[i][1] += scale*g[i][1];
	  v[i][2] += scale*g[i][2];
	}
	step /= 2;
      }
      else
	break;
    }
    if (fabs(rms_sq - rms_sq_last) < 1.e-14)
      break;
  }
}

static PyObject *
reduce_deformation_py(PyObject *dummy, PyObject *args)
{
  PyArrayObject *configuration;
  PyArrayObject *displacement;
  PyNonbondedListObject *nblist;
  double fc_length, factor;
  int niter;

  int natoms;
  vector3 *x, *v;

  vector3 *g;

  if (!PyArg_ParseTuple(args, "O!O!O!ddi",
			&PyArray_Type, &configuration,
			&PyArray_Type, &displacement,
			&PyNonbondedList_Type, &nblist,
			&fc_length, &factor, &niter))
    return NULL;
  natoms = configuration->dimensions[0];
  x = (vector3 *)configuration->data;
  v = (vector3 *)displacement->data;

  g = (vector3 *)malloc(natoms*sizeof(vector3));
  if (g == NULL) {
    PyErr_NoMemory();
    return NULL;
  }
  reduce_deformation(x, v, g, natoms, niter, nblist, fc_length, factor);
  free(g);

  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *
reduce_finite_deformation_py(PyObject *dummy, PyObject *args)
{
  PyArrayObject *configuration;
  PyArrayObject *displacement;
  PyNonbondedListObject *nblist;
  double fc_length, factor;
  double rms_reduction;

  int natoms;
  vector3 *x, *v;

  vector3 *g;

  if (!PyArg_ParseTuple(args, "O!O!O!ddd",
			&PyArray_Type, &configuration,
			&PyArray_Type, &displacement,
			&PyNonbondedList_Type, &nblist,
			&fc_length, &factor, &rms_reduction))
    return NULL;
  natoms = configuration->dimensions[0];
  x = (vector3 *)configuration->data;
  v = (vector3 *)displacement->data;

  g = (vector3 *)malloc(natoms*sizeof(vector3));
  if (g == NULL) {
    PyErr_NoMemory();
    return NULL;
  }
  reduce_finite_deformation(x, v, g, natoms, rms_reduction,
			    nblist, fc_length, factor);
  free(g);

  Py_INCREF(Py_None);
  return Py_None;
}


/* Forcefield evaluator for deformation energy */

static void
add_pair_term(PyFFEvaluatorObject *eval, PyObject *fc, fc_function *fc_fn,
	      int i, int j, vector3 dr, double r_sq, double f1, double f2)
{
  if (fc_fn != NULL) {
    if ((*fc_fn)(eval, fc, i, j, NULL, r_sq)) {
      tensor3 fij;
      int k, l;
      for (k = 0; k < 3; k++) {
	for (l = 0; l < 3; l++)
	  fij[k][l] = (f2-f1)*dr[k]*dr[l]/r_sq;
	fij[k][k] += f1;
      }
      (*fc_fn)(eval, fc, i, i, fij, r_sq);
      (*fc_fn)(eval, fc, j, j, fij, r_sq);
      tensor_changesign(fij);
      if (i > j)
	(*fc_fn)(eval, fc, j, i, fij, r_sq);
      else
	(*fc_fn)(eval, fc, i, j, fij, r_sq);
    }
  }
  else {
    double *data = (double *)((PyArrayObject *)fc)->data;
    int n = ((PyArrayObject *)fc)->dimensions[0];
    double *fcii = data + 9*n*i+3*i;
    double *fcjj = data + 9*n*j+3*j;
    double *fcij;
    int k, l;
    if (i > j) {
      int temp = i;  i = j;  j = temp;
    }
    fcij = data + 9*n*i+3*j;
    for (k = 0; k < 3; k++) {
      for (l = 0; l < 3; l++) {
	double f = (f2-f1)*dr[k]*dr[l]/r_sq;
	fcii[3*n*k+l] += f;
	fcjj[3*n*k+l] += f;
	fcij[3*n*k+l] -= f;
      }
      fcii[k*(3*n+1)] += f1;
      fcjj[k*(3*n+1)] += f1;
      fcij[k*(3*n+1)] -= f1;
    }
  }
}

double
deformation_evaluator(PyObject *object, int natoms,
		      PyArrayObject *coordinates,
		      PyObject *gradients, PyObject *force_constants,
		      gradient_function *gradient_fn, fc_function *fc_fn,
		      int small_change)
{
  PyFFEvaluatorObject *self = (PyFFEvaluatorObject *)object;
  vector3 *x = (vector3 *)coordinates->data;
  distance_fn *d_fn = self->universe_spec->distance_function;
  double *distance_data = self->universe_spec->geometry_data;

  PyNonbondedListObject *nblist = (PyNonbondedListObject *)self->data1;
  double fc_length = self->param1;
  double cutoff_sq = sqr(self->param2);
  double scale_factor = self->param3;
  double one_four_factor = self->param4;
  atom_pair *pair = nblist->pairs;
  int k;

  self->energy = 0.;
  self->virial = 0.;
  self->virial_available = 1;
  for (k = 0; k < nblist->npairs; k++, pair++) {
    if (cutoff_sq == 0. || pair->r_sq <= cutoff_sq) {
      int i = pair->i1;
      int j = pair->i2;
      double r_sq = pair->r_sq;
      vector3 rij;
      (*d_fn)(rij, x[j], x[i], distance_data);
      if (force_constants != NULL) {
        double deriv = 0.;
	double deriv2 = scale_factor * exp((0.01-r_sq)/sqr(fc_length));
	if (pair->one_four)
	  deriv2 *= one_four_factor;
	add_pair_term(self, force_constants, fc_fn, i, j,
		      rij, r_sq, deriv, deriv2);
      }
    }
  }
  return self->energy;
}

static PyObject *
DeformationEvaluator(PyObject *dummy, PyObject *args)
{
  PyFFEvaluatorObject *self = PyFFEvaluator_New();
  if (self == NULL)
    return NULL;
  if (!PyArg_ParseTuple(args, "O!Odddd",
			&PyUniverseSpec_Type, &self->universe_spec,
			&self->data1, &self->param1,
			&self->param2, &self->param3, &self->param4))
    return NULL;
  Py_INCREF(self->universe_spec);
  Py_INCREF(self->data1);
  self->eval_func = deformation_evaluator;
  self->name = "deformation";
  return (PyObject *)self;
}

double
calpha_evaluator(PyObject *object, int natoms,
		 PyArrayObject *coordinates,
		 PyObject *gradients, PyObject *force_constants,
		 gradient_function *gradient_fn, fc_function *fc_fn,
		 int small_change)
{
  PyFFEvaluatorObject *self = (PyFFEvaluatorObject *)object;
  vector3 *x = (vector3 *)coordinates->data;
  distance_fn *d_fn = self->universe_spec->distance_function;
  double *distance_data = self->universe_spec->geometry_data;

  PyNonbondedListObject *nblist = (PyNonbondedListObject *)self->data1;
  double cutoff_sq = sqr(self->param1);
  atom_pair *pair = nblist->pairs;
  int k;

  self->energy = 0.;
  self->virial = 0.;
  self->virial_available = 1;
  for (k = 0; k < nblist->npairs; k++, pair++) {
    if (cutoff_sq == 0. || pair->r_sq <= cutoff_sq) {
      int i = pair->i1;
      int j = pair->i2;
      double r_sq = pair->r_sq;
      vector3 rij;
      (*d_fn)(rij, x[j], x[i], distance_data);
      if (force_constants != NULL) {
        double deriv = 0.;
	double deriv2;
	if (r_sq < 0.1576)
#if 0
	  deriv2 = 400000.*r_sq;
#endif
#if 1
	  deriv2 = 1000000.*r_sq;
#endif
	else {
	  double r = sqrt(r_sq);
#if 0
	  deriv2 = 200./(r*r_sq);
#endif
#if 1
	  deriv2 = 500./(r*r_sq);
#endif
	}
	add_pair_term(self, force_constants, fc_fn, i, j,
		      rij, r_sq, deriv, deriv2);
      }
    }
  }
  return self->energy;
}

static PyObject *
CalphaEvaluator(PyObject *dummy, PyObject *args)
{
  PyFFEvaluatorObject *self = PyFFEvaluator_New();
  if (self == NULL)
    return NULL;
  if (!PyArg_ParseTuple(args, "O!Od",
			&PyUniverseSpec_Type, &self->universe_spec,
			&self->data1, &self->param1))
    return NULL;
  Py_INCREF(self->universe_spec);
  Py_INCREF(self->data1);
  self->eval_func = calpha_evaluator;
  self->name = "calpha_deformation";
  return (PyObject *)self;
}

/*
 * List of functions defined in the module
 */

static PyMethodDef deformation_methods[] = {
  {"deformation", deformation_py, 1},
  {"reduceDeformation", reduce_deformation_py, 1},
  {"reduceFiniteDeformation", reduce_finite_deformation_py, 1},
  {"DeformationEvaluator", DeformationEvaluator, 1},
  {"CalphaEvaluator", CalphaEvaluator, 1},
  {NULL, NULL}		/* sentinel */
};

/* Initialization function for the module */

void
initMMTK_deformation()
{
  PyObject *m;
  PyObject *module;

  /* Create the module and add the functions */
  m = Py_InitModule("MMTK_deformation", deformation_methods);

#ifdef import_array
  import_array();
#endif

  /* Import other modules and retrieve C API address arrays */
  module = PyImport_ImportModule("MMTK_forcefield");
  if (module != NULL) {
    PyObject *module_dict = PyModule_GetDict(module);
    PyObject *c_api_object = PyDict_GetItemString(module_dict, "_C_API");
    if (PyCObject_Check(c_api_object)) {
      PyFF_API = (void **)PyCObject_AsVoidPtr(c_api_object);
    }
  }
  module = PyImport_ImportModule("MMTK_universe");
  if (module != NULL) {
    PyObject *module_dict = PyModule_GetDict(module);
    PyObject *c_api_object = PyDict_GetItemString(module_dict, "_C_API");
    if (PyCObject_Check(c_api_object))
      PyUniverse_API = (void **)PyCObject_AsVoidPtr(c_api_object);
  }

  /* Check for errors */
  if (PyErr_Occurred())
    Py_FatalError("can't initialize module MMTK_deformation");
}
