/* Molecular surface calculations via NSC.
 *
 * Written by Lutz Ehrlich
 * Modified by Konrad Hinsen
 * last revision: 1999-2-25
 */

#include "forcefield.h"


/* Global variables */


static PyObject *
getContacts(PyObject *dummy, PyObject *args)
{
  PyArrayObject *configuration;
  PyObject *sigma1_list;
  float        *sigma1_arr;

  PyObject *sigma2_list;
  float        *sigma2_arr;

  PyObject *index1_list;
  int          *index1_arr;
  int           i1len;

  PyObject *index2_list;
  int          *index2_arr;
  int           i2len;

  PyObject *resultList;
  PyObject *tmpList;

  double contactFactor;
  int steps;
  char *description;
  vector3 *x, *v, *f;
  double *m;
  float dist, maxDist;
  double factor;
  int atoms;
  int ii, jj, kk;
  int index_ii, index_jj;
  long l_index_ii, l_index_jj, l_ii, l_jj;
  short  *con_arr;
  int     con_arr_ii;
  PyObject *obj; /* dummy object */
  
  


  /* Parse and check arguments */
  if (!PyArg_ParseTuple(args, "O!dO!O!O!O!",
			&PyArray_Type, &configuration,
			&contactFactor, 
			&PyList_Type, &sigma1_list,
			&PyList_Type, &index1_list,
			&PyList_Type, &sigma2_list,
			&PyList_Type, &index2_list))
    return NULL;

  /* Set some convenient variables */
  atoms = configuration->dimensions[0];
  x = (vector3 *)configuration->data;

  /* Python lists -> C arrays */
  i1len      = (int)    PyObject_Length((PyObject*)sigma1_list);
  sigma1_arr = (float*) malloc( i1len * sizeof(float));
  index1_arr = (int*)    malloc( i1len * sizeof(float));
  if( (!sigma1_arr) || (!index1_arr) ){
    return NULL;
  }

  i2len      = (int)    PyObject_Length((PyObject*)sigma2_list);
  sigma2_arr = (float*) malloc( i2len * sizeof(float));
  index2_arr = (int*)    malloc( i2len * sizeof(float));
  if( (!sigma2_arr) || (!index2_arr) ){
    return NULL;
  }

#ifdef DBG
  printf("allocation\n");
#endif
  
  for (ii = 0; ii < i1len ; ii++ ){
    obj            = PyList_GetItem((PyObject*)index1_list,ii);
    index1_arr[ii] = (int)   PyInt_AsLong(obj);
    obj            = PyList_GetItem((PyObject*)sigma1_list,ii);
    sigma1_arr[ii] = (float) PyFloat_AsDouble(obj);
  }
  for (ii = 0; ii < i2len ; ii++ ){
    obj            = PyList_GetItem((PyObject*)index2_list,ii);
    index2_arr[ii] = (int)   PyInt_AsLong(obj);
    obj            = PyList_GetItem((PyObject*)sigma2_list,ii);
    sigma2_arr[ii] = (float) PyFloat_AsDouble(obj);
  }
  
#ifdef DBG
  printf("arrays filled\n");
#endif
  
  /* create a C array for temporary storage of contacts */
  con_arr = (short*) calloc( i1len * i2len, sizeof( short ));
#ifdef DBG
  printf("con_arr init\n");
#endif
  
  /* do the N^2 operation */
  for ( ii = 0 ; ii < i1len ; ii++ ){
    index_ii = index1_arr[ii];
#ifdef DBG
  printf("(%d) x: %f %f %f \n", index_ii, x[index_ii][0], x[index_ii][1], x[index_ii][2]);
#endif

    for( jj = 0 ; jj < i2len ; jj++ ){
      index_jj = index2_arr[jj];
      dist = 0.;
      for ( kk = 0 ; kk < 3; kk++ ){
	dist += ( (x[index_ii][kk] - x[index_jj][kk]) *
		  (x[index_ii][kk] - x[index_jj][kk]) );
      }
      maxDist = contactFactor * (sigma1_arr[ii] + sigma2_arr[jj]);
      maxDist = maxDist * maxDist; /* avoid sqrt */
      if( dist <= maxDist ){
#ifdef DBG
  printf("dist:%f  maxDist:%f\n", dist, maxDist);
#endif     
  	
	con_arr[ ii * i2len + jj ] = 1;
      }
    }/* jj */
  }/* ii */
#ifdef DBG
  printf("n*n done\n");
#endif     
  
  resultList = (PyObject*)PyList_New(0);
#ifdef DBG
  printf("empty result list alloc\n");
#endif     
#ifdef DBG
  for ( ii = 0 ; ii < i1len ; ii++ ){
    con_arr_ii = ii * i2len;
    for ( jj = 0 ; jj < i2len ; jj++ ){
      if( con_arr[ con_arr_ii + jj] ){
	printf("con_arr[%d] = %d\n", con_arr_ii + jj, con_arr[ con_arr_ii + jj]);
      }
    }
  }
#endif     


  for ( ii = 0 ; ii < i1len ; ii++ ){
    con_arr_ii = ii * i2len;
    for ( jj = 0 ; jj < i2len ; jj++ ){
      if( con_arr[ con_arr_ii + jj] ){
	/* we have to build a new list and append it to the result list */
	index_ii = index1_arr[ii];
	index_jj = index2_arr[jj];
	tmpList = (PyObject*)PyList_New(0);
	if( tmpList == NULL ){
	  PyErr_NoMemory();
	  return NULL;
	}
#ifdef DBG
  printf("tmpList alloc\n");
#endif 
        kk = 0;
        l_ii = ii; /* copy int to long */
        PyList_Append(tmpList, PyInt_FromLong(l_ii));
#ifdef DBG
  printf("tmpList set 0\n");
#endif
        kk = 1;
	l_jj = jj; /* copy int to long */
	PyList_Append(tmpList, PyInt_FromLong(l_jj));
#ifdef DBG
  printf("tmpList set 1\n");
#endif 
	PyList_Append((PyObject*)resultList, tmpList );
#ifdef DBG
  printf("tmpListappended\n");
#endif 
      }
    }
  }
	
#ifdef DBG
  printf("result filled\n");
#endif	

  free(con_arr);
  free(sigma1_arr);
  free(index1_arr);
  free(sigma2_arr);
  free(index2_arr);

  return (PyObject*) resultList;

}


/* 
Routine for surface and volume computation using Frank Eisenhabers routine NSC.
For information on NSC, check:
   http://www.embl-heidelberg.de/Services/argos/ASC.21/nsc2-docu.html

NB: The academic license of NSC applies! You are not allowed to use this
    software in any commercial product.

ehr-25-7-97

*/


#include "nsc.h"

static PyObject *
getSurfaceAndVolume(PyObject *dummy, PyObject *args)
{
  PyArrayObject *configurationArr;
  PyArrayObject *radiiArr;
  PyObject *retList;


  char *description;
  int atoms;
  int ii, jj, kk;
  int ok;

  double *xyz;
  double *radii;
  double surface, volume;
  
  const int numPoints=700; /* set to value proposed in NSC docs */

  int NSC(double*, double*, int, int, int, double*, double**, double*,
	  double**,int*);
  

  /* Parse and check arguments */
  if (!PyArg_ParseTuple(args, "O!O!",
			&PyArray_Type, &configurationArr,
			&PyArray_Type, &radiiArr))
    return NULL;


  /* Set some convenient variables */
  atoms = radiiArr->dimensions[0];
  xyz   = (double *) configurationArr->data;
  radii = (double *) radiiArr->data;
  if ( 3 * atoms != (configurationArr->dimensions[0]) ){
    PyErr_SetString(PyExc_ValueError, "inconsistent input dimensions");
    return NULL;
  }

#ifdef DBG  
  for ( ii=0; ii< atoms; ii++ ){
    printf("%d: %f %f %f, rad: %f\n", ii, xyz[3 * ii], xyz[3 * ii + 1], xyz[3*ii+2], radii[ii]);
  }
#endif

  
  if( NSC(xyz,radii,atoms,numPoints,FLAG_VOLUME,
	  &surface,NULL, &volume,NULL,NULL) ){
    PyErr_SetString(PyExc_ValueError, "NSC error");
    return NULL;
  }
  
  retList = PyList_New(0);
  if (retList != NULL) {
    ok = (PyList_Append(retList, PyFloat_FromDouble(surface)) == 0) &&
         (PyList_Append(retList, PyFloat_FromDouble(volume)) == 0);
  
    if (!ok) {
      Py_DECREF(retList);
      retList = NULL;
    }
  }
  return (PyObject*) retList;

}

static PyObject *
getSurfaceAtoms(PyObject *dummy, PyObject *args)
{
  PyArrayObject *configurationArr;
  PyArrayObject *radiiArr;
  PyObject *areaList;


  char *description;
  int atoms;
  int ii, jj, kk;
  int ok;

  double *xyz;
  double *radii;
  double *areaArr;
  double surface;
  
  const int numPoints=700; /* set to value proposed in NSC docs */

  int NSC(double*, double*, int, int, int, double*, double**, double*,
	  double**,int*);
  

  /* Parse and check arguments */
  if (!PyArg_ParseTuple(args, "O!O!",
			&PyArray_Type, &configurationArr,
			&PyArray_Type, &radiiArr))
    return NULL;

  /* Set some convenient variables */
  atoms = radiiArr->dimensions[0];
  xyz   = (double *) configurationArr->data;
  radii = (double *) radiiArr->data;
  if ( 3 * atoms != (configurationArr->dimensions[0]) ){
    PyErr_SetString(PyExc_ValueError, "inconsistent input dimensions");
    return NULL;
  }

#ifdef DBG  
  for ( ii=0; ii< atoms; ii++ ){
    printf("%d: %f %f %f, rad: %f\n", ii, xyz[3 * ii], xyz[3 * ii + 1], xyz[3*ii+2], radii[ii]);
  }
#endif

  
  if( NSC(xyz,radii,atoms,numPoints,FLAG_ATOM_AREA,
	  &surface,&areaArr, NULL,NULL,NULL) ){
    PyErr_SetString(PyExc_ValueError, "NSC error");
    return NULL;
  }
  
  areaList = PyList_New(0);
  if (areaList != NULL) {
    ok = 1;
    for( ii=0; ii< atoms && ok; ii++ ){
      ok &= (PyList_Append( areaList, PyFloat_FromDouble( areaArr[ii] ) ) == 0);
    }
    if (!ok) {
      Py_DECREF(areaList);
      areaList = NULL;
    }
  }
  free(areaArr);

  return (PyObject*) areaList;

}

/*
 * List of functions defined in the module
 */

static PyMethodDef surface_methods[] = {
  {"getContacts", getContacts, 1},
  {"getSurfaceAndVolume", getSurfaceAndVolume, 1},
  {"getSurfaceAtoms", getSurfaceAtoms, 1},
  {NULL, NULL}		/* sentinel */
};

/* Initialization function for the module */

void
initMMTK_surface()
{
  PyObject *m;

  /* Create the module and add the functions */
  m = Py_InitModule("MMTK_surface", surface_methods);
  
  /* Import the array module */
#ifdef import_array
  import_array();
#endif

  /* Check for errors */
  if (PyErr_Occurred())
    Py_FatalError("can't initialize module MMTK_surface");
}
