# This module implements a class that does triangulation (tetrahedralization?)
# (mainly an interface to Qhull or Hull)
#
# Written by Peter McCluskey (pcm@rahul.net)
#

from Geometry import Polygon, Sphere, LineSegment, Prism, Plane, Pyramid
# need either hull or _qhull for this module to be of some use
try:
    import hull
except ImportError: pass
try:
    import _qhull
except ImportError: pass
import Units, Utility
from DictWithDefault import DictWithDefault
import pprint

#
# DelaunayTriangulation should not be created directly with convex_hull = 1;
# create a ConvexHull instead
#

class DelaunayTriangulation:

    def __init__(self, obj, danglingPositionsOnly = 0, qhull = 1,\
		 convex_hull = 0):
	self.qhull = qhull
	if qhull:
	    self.scale_factor = 1
	    self.min_dist = 1000
	else:
	    self.scale_factor = 1000
	    self.min_dist = 1000
	pos_list = []
	a_list = []
	self.pos_list = pos_list
	self.a_list = a_list
	pos_map = {}
	coincide_map = DictWithDefault([])
	self.coincide_map = coincide_map
	if danglingPositionsOnly:
	    for db in obj.danglingBondList():
		pos = db.DanglingPosition()
		if Utility.isDefinedPosition(pos):
		    pkey = self.pos_key(pos)
		    if pos_map.has_key(pkey):
			coincide_map[pos_map[pkey]].append(db)
		    else:
			pos_list.append(pos)
			a_list.append(db)
			pos_map[pkey] = len(a_list)-1
		    print 'DelaunayTri',len(pos_list)-1, Utility.isDefinedPosition(pos),db.a1.fullName(), self.pos_key(pos)
		    import sys
		    sys.stdout.flush()
	else:
	    try:
		atom_list = obj.atomList()
	    except AttributeError:
		atom_list = obj
	    for a in atom_list:
		pos = a.position()
		if Utility.isDefinedPosition(pos):
		    pkey = self.pos_key(pos)
		    if pos_map.has_key(pkey):
			print a.fullName(),'coincides with',\
			      a_list[pos_map[pkey]].fullName(),pkey
			coincide_map[pos_map[pkey]].append(a)
		    else:
			a_list.append(a)
			pos_list.append(pos)
			pos_map[pkey] = len(a_list)-1
			#print pos_map[pkey],'add',a.fullName(),pkey
	assert(len(pos_list) > 0)
	self.convex_hull = convex_hull
	if convex_hull:
	    print 'len(pos_list)',len(pos_list)
	    self.triang_index = hull.ConvexHull(pos_list, self.scale_factor)
	elif qhull:
	    self.triang_index = _qhull.Delaunay(pos_list)
	else:
	    self.triang_index = hull.Delaunay(pos_list, self.scale_factor)
	self.removed = DictWithDefault(0)
	self.faces_by_tet = []
	for tetrahedron in self.triang_index:
	    if -1 in tetrahedron:
		face = list(tetrahedron)
		face.remove(-1)
		self.faces_by_tet.append((tuple(face), ))
		continue
	    if convex_hull:
		self.faces_by_tet.append((tetrahedron, ))
		continue
	    faces_in_tet = []
	    for i in range(4):
		for j in range(i + 1, 4):
		    for k in range(j + 1, 4):
			faces_in_tet.append((tetrahedron[i], tetrahedron[j],\
					     tetrahedron[k]))
	    self.faces_by_tet.append(faces_in_tet)

    def __del__(self):
	if self.qhull:
	    _qhull.__del__()

    def numberOfTetrahedra(self):
	return len(self.triang_index)

    def coordList(self, start_index = 0, stop_index = -1):
	if start_index < 0:
	    start_index = len(self.pos_list) + start_index
	if stop_index < 0:
	    stop_index = len(self.pos_list) + stop_index
	coords = []
	if 1:
	    for i in range(len(self.triang_index)):
		in_range = 0
		for v_index in self.triang_index[i]:
		    if v_index >= start_index and v_index < stop_index:
			in_range = 1
			break
		if in_range:
		    coords.append(self.tetrahedronCoordList(i))
	    return coords
	edges = self.listEdges()
	for e in edges:
	    if (e[0] >= start_index or e[1] >= start_index) and \
	       (e[0] <= stop_index or e[1] <= stop_index):
		coords.append([self.pos_list[e[0]], self.pos_list[e[1]]])
	return coords

    def tetrahedronCoordList(self, index):
	if self.removed[index]:
	    return []
	t = self.triang_index[index]
	coords = []
	for i in (t[0], t[1], t[2], t[0], t[3], t[1], t[3], t[2]):
	    if i != -1:
		coords.append(self.pos_list[i])
	return coords

    def buildNeighborDict(self):
	neighbors = DictWithDefault([])
	used = {}
	for tet in self.triang_index:
	    for i in (0, 1, 2, 3):
		for j in range(i+1, 4):
		    if tet[i] == -1 or tet[j] == -1:
			continue	# make this optional?
		    if not used.has_key((tet[i], tet[j])) and\
		       not used.has_key((tet[j], tet[i])):
			used[(tet[i], tet[j])] = 1
			neighbors[tet[i]].append(tet[j])
			neighbors[tet[j]].append(tet[i])
	self.neighbors = neighbors
	return neighbors

    def removeFacesBiggerThan(self, area, radius = None):
	if radius is not None and not hasattr(self, 'neighbors'):
	    self.buildNeighborDict()
	face_outside = DictWithDefault(0)
	tet_outside = DictWithDefault(0)
	is_inside = DictWithDefault(0)
	num_removed = 0
	surface_tets = []
	tets_by_face = DictWithDefault([])
	for i in range(len(self.faces_by_tet)):
	    tetrahedron_faces = self.faces_by_tet[i]
	    if len(tetrahedron_faces) == 1:
		tet_outside[id(tetrahedron_faces)] = 1
		face_outside[face_key(tetrahedron_faces[0])] = 1
		#print 'face_outside',tetrahedron_faces[0]
		surface_tets.append(i)
	    else:
		is_inside[id(tetrahedron_faces)] = 1
		for face in tetrahedron_faces:
		    face_list = list(face)
		    face_list.sort()
		    tets_by_face[face_key(face)].append(i)
	while 1:
	    changed = 0
	    num_outside_faces = len(face_outside.keys())
	    new_surface_tets = []
	    for i in surface_tets:
		tetrahedron_faces = self.faces_by_tet[i]
		for face in tetrahedron_faces:
		    #print 'face_outside?',i,face_outside[face],face
		    if face_outside[face_key(face)]:
			triangle = Polygon([self.pos_list[face[0]],\
					    self.pos_list[face[1]],\
					    self.pos_list[face[2]]])
			face_area = triangle.area()
			if 0:
			    print 'removing?',i,face_area,triangle,\
			      '%.2f %.2f %.2f' % ((self.pos_list[face[0]] - self.pos_list[face[1]]).length(),\
						  (self.pos_list[face[1]] - self.pos_list[face[2]]).length(),\
						  (self.pos_list[face[2]] - self.pos_list[face[0]]).length())
			if face_area > area or (radius is not None and\
						self.fitsRadius(face, radius)):
			    self.removed[i] = 1
			    changed = changed + 1
			    num_removed = num_removed + 1
			    tet_outside[id(tetrahedron_faces)] = 1
			    for new_face in tetrahedron_faces:
				key = face_key(new_face)
				face_outside[key] = 1
				#print 'face',key,'made outside',len(tets_by_face[key])
				for tet_index in tets_by_face[key]:
				    tet = self.faces_by_tet[tet_index]
				    if is_inside[id(tet)]:
					new_surface_tets.append(tet_index)
					is_inside[id(tet)] = 0
			    break
	    if changed == 0:
		break
	    print 'pass done',changed, num_outside_faces, len(face_outside.keys()),len(new_surface_tets),len(self.faces_by_tet),len(tet_outside.keys())
	    surface_tets = new_surface_tets
	print num_removed,'tetrahedra removed of',len(self.triang_index)
	
	
    def fitsRadius(self, face, radius):
	triangle = Polygon([self.pos_list[face[0]],\
			    self.pos_list[face[1]],\
			    self.pos_list[face[2]]])
	edges = triangle.edgeVectors()
	sum = 0
	for edge in edges:
	    sum = sum + edge.length()
	    if edge.length() > 2*radius:
		break
	    #else: print 'edge short',face,'[',edges.index(edge),']',edge
	else:
	    return 0
	sphere1 = Sphere(triangle.center(), radius)
	for v in face:
	    for index in self.neighbors[v]:
		if index != -1 and sphere1.enclosesPoint(self.pos_list[index]):
		    return 0
	return 1

    def contains(self, point):
	if self.qhull:
	    index = _qhull.FindFacet(point)
	    if index >= 0 and not self.removed[index]:
		return index + 1
	    return 0
	for i in range(len(self.triang_index)):
	    if self.removed[i]:
		continue
	    tetrahedron = self.triang_index[i]
	    if not (-1 in tetrahedron):
		tri = Polygon([self.pos_list[tetrahedron[0]],\
			       self.pos_list[tetrahedron[1]],\
			       self.pos_list[tetrahedron[2]]])
		p = tri.projectionOf(point)
		if p is not None and (p - tri[0]) * tri.normal > 0:
		    return i+1
	return 0

    def listContains(self, points):
	pt_map = DictWithDefault([])
	fact = 25.0
	for i in range(len(points)):
	    pt = points[i]
	    key1 = (int(pt[0]*fact), int(pt[1]*fact), int(pt[2]*fact))
	    pt_map[key1].append(i)
	    if i == 3777 or i == 4346 or i == 3159:
		print i,key1,pt
	ret_list = []
	for i in range(len(self.triang_index)):
	    if self.removed[i]:
		continue
	    tetrahedron = self.triang_index[i]
	    if not (-1 in tetrahedron):
		tet = Pyramid([self.pos_list[tetrahedron[0]],\
			       self.pos_list[tetrahedron[1]],\
			       self.pos_list[tetrahedron[2]],\
			       self.pos_list[tetrahedron[3]]])
		faces = [(3, Plane(self.pos_list[tetrahedron[0]],\
			       self.pos_list[tetrahedron[1]],\
			       self.pos_list[tetrahedron[2]])),\
			 (0, Plane(self.pos_list[tetrahedron[1]],\
			       self.pos_list[tetrahedron[2]],\
			       self.pos_list[tetrahedron[3]])),\
			 (1, Plane(self.pos_list[tetrahedron[0]],\
			       self.pos_list[tetrahedron[2]],\
			       self.pos_list[tetrahedron[3]])),\
			 (2, Plane(self.pos_list[tetrahedron[0]],\
			       self.pos_list[tetrahedron[1]],\
			       self.pos_list[tetrahedron[3]]))]
		for pos_key in self._positionsNear(tet, fact):
		    for point_index in pt_map[pos_key]:
			point = points[point_index]
			c = 0
			for (out, tri) in faces:
			    p = tri.projectionOf(point)
			    out_pos = self.pos_list[tetrahedron[out]]
			    if p is not None and (out_pos - p) * (point - p) > 0:
				c = c + 1
			if point_index == 3777 or point_index == 4346 or point_index == 3159:
			    print c,i,'point_index',point_index,point,tet
			if c == 4:
			    ret_list.append(point_index)
	return ret_list

    def _positionsNear(self, tet, fact):
	(minv, maxv) = tet.boundingBox()
	list = []
	for x in range(int(minv[0]*fact)-1, int(maxv[0]*fact)+2):
	    for y in range(int(minv[1]*fact)-1, int(maxv[1]*fact)+2):
		for z in range(int(minv[2]*fact)-1, int(maxv[2]*fact)+2):
		    list.append((x, y, z))
	return list

    def pos_key(self, pos):
	return (int(pos[0]*self.min_dist), \
		int(pos[1]*self.min_dist), \
		int(pos[2]*self.min_dist))

    def findNearKey(self, index, pos_map):
	key = self.pos_key(self.pos_list[index])
	for x in range(key[0]-1, key[0]+2):
	    for y in range(key[1]-1, key[1]+2):
		for z in range(key[2]-1, key[2]+2):
		    if pos_map.has_key((x, y, z)):
			return (x, y, z)
	return None

    def listEdges(self, include_unbounded = 0, include_removed = 0):
	used = {}
	edge_list = []
	for index in range(len(self.triang_index)):
	    if self.removed[index]:
		continue
	    t = self.triang_index[index]
	    if len(t) != 4:
		msg = 'DelaunayTriangulation.listEdges invalid tet',t
		raise RuntimeError, msg
	    for i in range(4):
		ti = t[i]
		if ti == -1:
		    continue
		for j in range(i+1, 4):
		    key = (ti, t[j])
		    if key[0] == key[1]:
			continue
		    if not include_unbounded and (key[0] == -1 or key[1] == -1):
			continue
		    if not used.has_key(key):
			used[key] = 1
			edge_list.append(key)
	return edge_list
		
    def listPositionsCloserThan(self, max_dist, group_size = 2):
	co_map = self.coincide_map
	edge_list = self.listEdges()
	near_pos = DictWithDefault([])
	done_already = {}
	plist = []
	for edge in edge_list:
	    #print 'pos_dist %.3f' % (self.pos_list[edge[0]] - self.pos_list[edge[1]]).length(), self.a_list[edge[0]], self.a_list[edge[1]]
	    if (self.pos_list[edge[0]] - self.pos_list[edge[1]]).length() < max_dist:
		if group_size == 2:
		    pair = (self.a_list[edge[0]], self.a_list[edge[1]])
		    if not done_already.has_key(pair):
			done_already[pair] = 1
			done_already[(pair[1], pair[0])] = 1
			plist.append(pair)
		else:
		    near_pos[edge[0]].append(self.a_list[edge[1]])
		    near_pos[edge[1]].append(self.a_list[edge[0]])
	for (index, altindex_list) in co_map.items():
	    for ai in altindex_list:
		near_pos[index].append(ai)
	if group_size > 2:
	    for (key, edge_list) in near_pos.items():
		if len(edge_list) >= group_size - 1:
		    plist.append([self.a_list[key]] + edge_list)
	return plist
	
def face_key(face):
    face_list = list(face)
    face_list.sort()
    return tuple(face_list)

class ConvexHull(DelaunayTriangulation):
    def __init__(self, obj):
	DelaunayTriangulation.__init__(self, obj, convex_hull = 1, qhull = 0)

    def nearestPointOnSurface(self, atom_pos, path = None):
	best_point = None
	best_dist = 1.e9
	tri_list = []
	for triangle in self.triang_index:
	    pt_list = []
	    tri_name = ''
	    for index in triangle:
		pt_list.append(self.a_list[index].position())
		tri_name = tri_name + self.a_list[index].fullName()[-4:] + ','
	    poly = Polygon(pt_list)
	    tri_list.append(poly)
	    near_pt = poly.pointNearest(atom_pos)
	    d = (near_pt - atom_pos).length()
	    if d < best_dist:
		best_dist = d
		best_point = near_pt
		#print 'best_dist',d,tri_name,len(tri_list)
	if path is not None and best_point is not None:
	    line_seg = LineSegment(atom_pos, best_point)
	    for poly in tri_list:
		if poly.intersectWith(line_seg):
		    path.append(poly)
	    sort_fn = lambda t1, t2, pos = atom_pos : \
		      cmp((t1.center() - pos).length(), \
			  (t2.center() - pos).length())
	    path.sort(sort_fn)
	return best_point
