
#include "Python.h"
#include "qhull_a.h"

char qh_version[] = "user_eg 98/7/31 Python interface";  /* used for error messages */

static coordT *points = NULL; /* array of coordinates for each point */
static int *index_of_vextex = NULL;
static int *facet_index = NULL;
static int *is_on_hull = NULL;
static int curlong, totlong;   /* memory remaining after qh_memfreeshort */
static int keep_data = True;
static int n_dimensions = -1;

static PyObject *
make_py_num(long i)
{
  PyObject* numberi = PyInt_FromLong(i);
  Py_DECREF(numberi);
  return numberi;
}

static void
mark_extremes_d(facetT *facetlist, setT *facets, int printall, int *hull)
{
  setT *vertices;
  vertexT *vertex, **vertexp;
  boolT upperseen, lowerseen;
  facetT *neighbor, **neighborp;
  int numpoints=0;

  vertices= qh_facetvertices (facetlist, facets, printall);
  qh_vertexneighbors();
  FOREACHvertex_(vertices) {
    upperseen= lowerseen= False;
    FOREACHneighbor_(vertex) {
      if (neighbor->upperdelaunay)
        upperseen= True;
      else
        lowerseen= True;
    }
    if (upperseen && lowerseen) {
      hull[vertex->id] = True;
    }else
      hull[vertex->id] = False;
  }
   qh_settempfree (&vertices);
}

static void free_data()
{
  qh_freeqhull(!qh_ALL);                 /* free long memory */
  qh_memfreeshort (&curlong, &totlong);  /* free short memory and memory allocator */
  if (curlong || totlong) 
    fprintf (stderr, "qhull internal warning (user_eg, #2): did not free %d bytes of long memory (%d pieces)\n", totlong, curlong);
  free(facet_index);
  facet_index = NULL;
  free(index_of_vextex);
  index_of_vextex = NULL;
  free(is_on_hull);
  is_on_hull = NULL;
  free(points);
  points = NULL;
}

static PyObject *
Delaunay(PyObject *dummy, PyObject *args)
{
	int dim;	            /* dimension of points */
	int numpoints;            /* number of points */
	boolT ismalloc= False;    /* True if qhull should free points in qh_freeqhull() or reallocation */
	char flags[250];          /* option flags for qhull, see qh_opt.htm */
	int exitcode;             /* 0 if no error from qhull */
	facetT *facet;	    /* set by FORALLfacets */
	int i;
	int j;
	PyObject* py_points;
	PyObject* py;
	PyObject* result_list = NULL;
	if (!PyArg_ParseTuple(args, "O", &py_points))
	  return NULL;
	numpoints = PyObject_Length(py_points);
	if(numpoints <= 0)
	{
	  PyErr_SetString(PyExc_ValueError,
			  "len(py_points) didn't return positive number");
	  return NULL;
	}
	py = PyObject_GetItem(py_points, make_py_num(0));
	if(!py)
	{
	  PyErr_SetString(PyExc_IndexError, "py_points[0] failed");
	  return NULL;
	}
	dim = PyObject_Length(py);
	if (dim < 2)
	{
	  char buf[512];
	  sprintf(buf, "Number of dimensions (%d) less than 2", dim);
	  PyErr_SetString(PyExc_ValueError, buf);
	  return NULL;
	}

	result_list = PyList_New(0);

	points = (coordT*)malloc(sizeof(coordT)*(dim+1)*numpoints);
	for(j = 0; j < numpoints; ++j)
	{
	  py = PyObject_GetItem(py_points, make_py_num(j));
	  if(!py)
	  {
	    char buf[512];
	    sprintf(buf, "py_points[%d] failed", j);
	    PyErr_SetString(PyExc_IndexError, buf);
	    goto error;
	  }
	  for(i = 0; i < dim; ++i)
	  {
	    PyObject *v = PyObject_GetItem(py, make_py_num(i));
	    if(!v)
	    {
	      char buf[512];
	      sprintf(buf, "py_points[%d][%d] failed", j, i);
	      PyErr_SetString(PyExc_IndexError, buf);
	      goto error;
	    }
	    points[j*dim + i] = PyFloat_AsDouble(v);
	  }
	}

  sprintf (flags, "qhull s d QJ Tcv %s", "");
  exitcode= qh_new_qhull (dim, numpoints, points, ismalloc,
			  flags, NULL, stderr); 
  if(0) fprintf(stdout, "dim %d qh hull_dim %d\n", dim, qh hull_dim);
  if (!exitcode) {                  /* if no error */
    /* 'qh facet_list' contains the convex hull */
    qh_setvoronoi_all();
    exitcode= setjmp (qh errexit);  
    if (!exitcode) {
      /* Trap Qhull errors.  Without the setjmp(), Qhull
         will exit() after reporting an error */
      unsigned max_ids = qh vertex_id;
      unsigned max_facets = qh facet_id;
      int facet_no = 0;
      index_of_vextex = (int*)malloc(sizeof(int)*max_ids);
      facet_index = (int*)malloc(sizeof(int)*max_facets);
      is_on_hull = (int*)malloc(sizeof(int)*max_ids);
      qh NOerrexit= False;
      for(j = 0; j < max_facets; ++j)
	facet_index[j] = -2;
      for(j = 0; j < max_ids; ++j)
      {
	index_of_vextex[j] = -2;
	is_on_hull[j] = False;
      }
      mark_extremes_d(qh facet_list, NULL, 0, is_on_hull);
      for(j = 0; j < numpoints; ++j)
      {
	boolT isoutside;
	realT bestdist;
	facetT *facet;
	vertexT *vertex, **vertexp;
	coordT point1[100];
	int k;
	for(k = 0; k < dim; ++k)
	  point1[k] = points[k + j*dim];
	
	qh_setdelaunay (dim+1, 1, point1);
	facet= qh_findbestfacet (point1, qh_ALL, &bestdist, &isoutside);
	vertex = qh_nearvertex(facet, point1, &bestdist);
	index_of_vextex[vertex->id] = j;
	if(0)
	{
	  printf("%2d -> %2d ",j, vertex->id);
	  for (k=0; k < dim; k++)
	    printf ("%5.2f ", points[k + j*dim]);
	  for (k=0; k < dim; k++)
	    printf ("%5.2f ", vertex->point[k]);
	  printf ("\n");
	}
      }
      FORALLfacets {
	vertexT *vertex, **vertexp;
	int count_on_hull = 0;
	boolT not_used = False;
	PyObject* tuple;
	FOREACHvertex_(facet->vertices) {
	  if(index_of_vextex[vertex->id] == -2)
	  {
	    printf("%d (%d) not used %6.2f %6.2f %6.2f\n",
		   index_of_vextex[vertex->id], vertex->id,
		   vertex->point[0],vertex->point[1],vertex->point[2]);
	    not_used = True;
	  }
	}
	if(not_used)
	  continue;
	if(facet->upperdelaunay)
	{
	  facet_index[facet->id] = -1;
	  continue;
	}
	facet_index[facet->id] = facet_no++;
	tuple = PyTuple_New(dim+1);
	if(tuple)
	{
	  j = 0;
	  FOREACHvertex_(facet->vertices) {
	    if(0) printf("%3u(%d) ",index_of_vextex[vertex->id],vertex->id);
	    PyTuple_SET_ITEM(tuple, j, PyInt_FromLong(index_of_vextex[vertex->id]));
	    if(is_on_hull[vertex->id])
	      ++count_on_hull;
	    ++j;
	  }
	  if(0) printf("\n");
	  PyList_Append(result_list, tuple);
	}
	if(0 && count_on_hull > dim)
	{
	  FOREACHvertex_(facet->vertices) {
	    if(1) printf("%3u(%d) ",index_of_vextex[vertex->id],vertex->id);
	  }
	  if(1) printf("is hull\n");
	}
	if(0 && count_on_hull == dim)
	{
	  PyObject* tuple = PyTuple_New(dim+1);
	  if(tuple)
	  {
	    j = 0;
	    FOREACHvertex_(facet->vertices) {
	      if(is_on_hull[vertex->id])
		PyTuple_SET_ITEM(tuple, j, PyInt_FromLong(index_of_vextex[vertex->id]));
	      else PyTuple_SET_ITEM(tuple, j, PyInt_FromLong(-1));
	      ++j;
	    }
	    PyList_Append(result_list, tuple);
	  }
	  if(0) printf(" *");
	}
	if(0) printf ("\n");
      }
    }
    qh NOerrexit= True;
  }
  n_dimensions = dim;
 error:
  if(!keep_data)
    free_data();
  return result_list;
}

static PyObject*
Destroy(PyObject *dummy, PyObject *dummy2)
{
  free_data();
  Py_INCREF(Py_None);
  return Py_None;
}

/*--------------------------------------------------
  assumes dim < 100
notes:
  calls qh_setdelaunay() to project the point to a parabaloid
*/
#define MAX_F_DIM 100
static PyObject*
FindFacet(PyObject *dummy, PyObject *args)
{
  int k;
  int dim = n_dimensions;
  coordT point[MAX_F_DIM];
  boolT isoutside;
  realT bestdist;
  facetT *facet;
  vertexT *vertex, **vertexp;
  PyObject* py_point;
  PyObject* result_list = NULL;
  if (!PyArg_ParseTuple(args, "O", &py_point))
    return NULL;
  if(dim >= MAX_F_DIM)
  {
    PyErr_SetString(PyExc_ValueError, "Number of dimensions too high");
    return NULL;
  }
  for (k= 0; k < dim; k++) 
  {
    PyObject *v = PyObject_GetItem(py_point, make_py_num(k));
    if(!v)
    {
      char buf[512];
      sprintf(buf, "point[%d] failed", k);
      PyErr_SetString(PyExc_IndexError, buf);
      return NULL;
    }
    point[k] = PyFloat_AsDouble(v);
  }
  qh_setdelaunay (dim+1, 1, point);
  facet= qh_findbestfacet (point, !qh_ALL, &bestdist, &isoutside);
  return PyLong_FromLong(facet_index[facet->id]);
}

/*
 * List of functions defined in the module
 */

static PyMethodDef qhull_methods[] = {
  {"Delaunay", Delaunay, METH_VARARGS},
  {"FindFacet", FindFacet, METH_VARARGS},
  {"__del__", Destroy, METH_VARARGS},
  {NULL, NULL}		/* sentinel */
};

/* Initialization function for the module */

void
init_qhull()
{
  PyObject *m;

  /* Create the module and add the functions */
  m = Py_InitModule("_qhull", qhull_methods);
  
  /* Check for errors */
  if (PyErr_Occurred())
    Py_FatalError("can't initialize module qhull");
}
